from .base import MyApiTestCase


class APITokengroupTestCase(MyApiTestCase):

    def test_01_set_get_delete_tokengroup(self):
        # Set tokengroup
        with self.app.test_request_context('/tokengroup/gruppe1',
                                           data={"description": "My Cool first group"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertGreaterEqual(value, 1)

        # get tokengroups
        with self.app.test_request_context('/tokengroup/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertIn("gruppe1", value)
            self.assertEqual(value["gruppe1"]["description"], "My Cool first group")

        # create a 2nd group
        with self.app.test_request_context('/tokengroup/gruppe2',
                                           data={"description": "2nd group"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertGreaterEqual(value, 1)

        # get all tokengroups
        with self.app.test_request_context('/tokengroup/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(len(value), 2)
            self.assertIn("gruppe1", value)
            self.assertIn("gruppe2", value)

        # Change the description of the first group
        with self.app.test_request_context('/tokengroup/gruppe1',
                                           data={"description": "1st group"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertGreaterEqual(value, 1)

        # get 1st group
        with self.app.test_request_context('/tokengroup/gruppe1',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(len(value), 1)
            self.assertEqual(value["gruppe1"]["description"], "1st group")

        # delete 1st group
        with self.app.test_request_context('/tokengroup/gruppe1',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(value, 1)

        # check, that only the 2nd group is available
        with self.app.test_request_context('/tokengroup/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(len(value), 1)
            self.assertNotIn("gruppe1", value)
            self.assertIn("gruppe2", value)

        # delete 2nd group
        with self.app.test_request_context('/tokengroup/gruppe2',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(value, 1)



"""
This tests the package lib.utils
"""
from privacyidea.config import TestingConfig
from .base import MyTestCase, OverrideConfigTestCase

from privacyidea.lib.utils import (parse_timelimit,
                                   check_time_in_range, parse_proxy,
                                   check_proxy, reduce_realms, is_true,
                                   parse_date, compare_condition,
                                   get_data_from_params, parse_legacy_time,
                                   int_to_hex, compare_value_value,
                                   compare_generic_condition,
                                   parse_time_offset_from_now, censor_connect_string,
                                   parse_timedelta, to_unicode,
                                   parse_int, convert_column_to_unicode,
                                   truncate_comma_list, check_pin_contents, CHARLIST_CONTENTPOLICY,
                                   get_module_class, decode_base32check,
                                   get_client_ip, sanity_name_check, to_utf8,
                                   to_byte_string, hexlify_and_unicode,
                                   b32encode_and_unicode, to_bytes,
                                   b64encode_and_unicode, create_img,
                                   convert_timestamp_to_utc, modhex_encode,
                                   modhex_decode, checksum, urlsafe_b64encode_and_unicode,
                                   check_ip_in_policy, split_pin_pass, create_tag_dict,
                                   check_serial_valid, determine_logged_in_userparams,
                                   to_list, parse_string_to_dict, convert_imagefile_to_dataimage,
                                   get_plugin_info_from_useragent, get_computer_name_from_user_agent)
from privacyidea.lib.crypto import generate_password
from datetime import timedelta, datetime
from netaddr import IPAddress, IPNetwork, AddrFormatError
from dateutil.tz import tzlocal, tzoffset, gettz
from privacyidea.lib.tokenclass import DATE_FORMAT
from privacyidea.lib.error import PolicyError
import binascii


class UtilsTestCase(MyTestCase):

    def test_01_timelimit(self):
        c, tdelta = parse_timelimit("1/5s")
        self.assertEqual(c, 1)
        self.assertEqual(tdelta, timedelta(seconds=5))

        c, tdelta = parse_timelimit("5/10M")
        self.assertEqual(c, 5)
        self.assertEqual(tdelta, timedelta(minutes=10))

        c, tdelta = parse_timelimit(" 5 / 10M ")
        self.assertEqual(c, 5)
        self.assertEqual(tdelta, timedelta(minutes=10))

        c, tdelta = parse_timelimit("7/120h")
        self.assertEqual(c, 7)
        self.assertEqual(tdelta, timedelta(hours=120))
        self.assertEqual(tdelta, timedelta(days=5))

        # A missing time specifier raises an Exception
        self.assertRaises(Exception, parse_timelimit, "7/12")

        # A non number raises an Exception
        self.assertRaises(Exception, parse_timelimit, "seven/12m")

    def test_02_timedelta(self):
        tdelta = parse_timedelta("123d")
        self.assertEqual(tdelta, timedelta(days=123))

        tdelta = parse_timedelta("31h")
        self.assertEqual(tdelta, timedelta(hours=31))

        tdelta = parse_timedelta(" 2y")
        self.assertEqual(tdelta, timedelta(days=2 * 365))

        tdelta = parse_timedelta("30 m ")
        self.assertEqual(tdelta, timedelta(minutes=30))

        # A missing time specifier raises an Exception
        self.assertRaises(Exception, parse_timedelta, "7")

        # A non number raises an Exception
        self.assertRaises(Exception, parse_timedelta, "sevenm")

    def test_03_check_time_in_range(self):
        # April 5th, 2016 is a Tuesday
        t = datetime(2016, 4, 5, hour=9, minute=12)
        r = check_time_in_range("Mon-Fri: 09:00-17:30", t)
        self.assertEqual(r, True)
        r = check_time_in_range("Mon - Fri : 09:00- 17:30", t)
        self.assertEqual(r, True)
        r = check_time_in_range("Sat-Sun:10:00-15:00, Mon - Fri : 09:00- "
                                "17:30", t)
        self.assertEqual(r, True)

        # Short time description
        r = check_time_in_range("Tue: 9-15", t)
        self.assertEqual(r, True)
        r = check_time_in_range("Tue: 9-15:1", t)
        self.assertEqual(r, True)

        # day out of range
        r = check_time_in_range("Wed - Fri: 09:00-17:30", t)
        self.assertEqual(r, False)

        # time out of range
        r = check_time_in_range("Mon-Fri: 09:30-17:30", t)
        self.assertEqual(r, False)

        # A time with a missing leading 0 matches anyway
        r = check_time_in_range("Mon-Fri: 9:12-17:30", t)
        self.assertEqual(r, True)

        # Nonsense will not match
        r = check_time_in_range("Mon-Wrong: asd-17:30", t)
        self.assertEqual(r, False)

    def test_04a_parse_proxy(self):
        self.assertEqual(parse_proxy(""), set())
        # 127.0.0.1 may rewrite to any IP
        self.assertEqual(parse_proxy("127.0.0.1"),
                         {(IPNetwork("127.0.0.1/32"), IPNetwork("0.0.0.0/0"))})
        # 127.0.0.x may rewrite to 10.0.x.x
        self.assertEqual(parse_proxy("127.0.0.1/24 >  10.0.0.0/16"),
                         {(IPNetwork("127.0.0.1/24"), IPNetwork("10.0.0.0/16"))})
        # 127.0.0.x may rewrite to 10.0.x.x or 10.1.x.x which may rewrite to 10.2.0.x
        self.assertEqual(parse_proxy("127.0.0.1/24>10.0.0.0/16, 127.0.0.1/24>10.1.0.0/16>10.2.0.0/24"),
                         {
                             (IPNetwork("127.0.0.1/24"), IPNetwork("10.0.0.0/16")),
                             (IPNetwork("127.0.0.1/24"), IPNetwork("10.1.0.0/16"), IPNetwork("10.2.0.0/24"))
                         })

    def test_04b_check_overrideclient(self):
        proxy_def = " 10.0.0.12, 1.2.3.4/16> 192.168.1.0/24, 172.16.0.1 " \
                    ">10.0.0.0/8   "
        r = parse_proxy(proxy_def)

        self.assertEqual(len(r), 3)
        self.assertIn((IPNetwork("1.2.3.4/16"), IPNetwork("192.168.1.0/24")), r)
        self.assertIn((IPNetwork("10.0.0.12/32"), IPNetwork("0.0.0.0/0")), r)
        self.assertIn((IPNetwork("172.16.0.1/32"), IPNetwork("10.0.0.0/8")), r)

        # check paths with only a single hop
        self.assertEqual(check_proxy(list(map(IPAddress, ["10.0.0.12", "1.2.3.4"])), proxy_def),
                         IPAddress("1.2.3.4"))  # 10.0.0.12 may map to 1.2.3.4
        self.assertEqual(check_proxy(list(map(IPAddress, ["10.0.0.11", "1.2.3.4"])), proxy_def),
                         IPAddress("10.0.0.11"))  # 10.0.0.11 may not map to 1.2.3.4
        self.assertEqual(check_proxy(list(map(IPAddress, ["1.2.3.10", "192.168.1.12"])), proxy_def),
                         IPAddress("192.168.1.12"))  # 1.2.3.10 may map to 192.168.1.12
        self.assertEqual(check_proxy(list(map(IPAddress, ["172.16.0.1", "1.2.3.4"])), proxy_def),
                         IPAddress("172.16.0.1"))  # 172.16.0.1 may not map to 1.2.3.4
        self.assertEqual(check_proxy(list(map(IPAddress, ["172.16.0.1", "10.1.2.3"])), proxy_def),
                         IPAddress("10.1.2.3"))  # 172.16.0.1 may map to 10.1.2.3

        # Wrong proxy setting. No commas (issue 526)
        proxy_def = " 10.0.0.12 1.2.3.4/16> 192.168.1.0/24 172.16.0.1 " \
                    ">10.0.0.0/8   "
        self.assertRaises(AddrFormatError, parse_proxy, proxy_def)
        # 10.0.0.12 is not allowed to map because the proxy settings are invalid
        self.assertEqual(check_proxy(list(map(IPAddress, ["10.0.0.12", "1.2.3.4"])), proxy_def),
                         IPAddress("10.0.0.12"))

        # check paths with several hops
        # 1.2.3.4 -------> 10.0.0.1 -------> 192.168.1.1 --------> privacyIDEA
        #  client           proxy1              proxy2
        path_to_client = list(map(IPAddress, ["192.168.1.1", "10.0.0.1", "1.2.3.4"]))
        # no proxy setting: client IP is proxy2
        self.assertEqual(check_proxy(path_to_client, ""),
                         IPAddress("192.168.1.1"))
        # proxy2 may map to 10.0.1.x: client IP is proxy2
        self.assertEqual(check_proxy(path_to_client, "192.168.1.1>10.0.1.0/24"),
                         IPAddress("192.168.1.1"))
        # proxy2 may map to 10.0.0.x: client IP is proxy1
        self.assertEqual(check_proxy(path_to_client, "192.168.1.1>10.0.0.0/24"),
                         IPAddress("10.0.0.1"))
        # proxy2 may map to 10.0.0.x, which may map to 2.3.4.x but not 1.2.3.4, so
        # the proxy definition does not match and the client IP is proxy2
        self.assertEqual(check_proxy(path_to_client, "192.168.1.1>10.0.0.0/24>2.3.4.0/24"),
                         IPAddress("192.168.1.1"))
        # 10.0.0.x may map to 2.3.4.x, but it doesn't matter because there is proxy2 inbetween
        self.assertEqual(check_proxy(path_to_client, "10.0.0.0/24>2.3.4.0/24"),
                         IPAddress("192.168.1.1"))
        # proxy2 may map to 10.0.0.x, which may map to 1.2.x.x or 2.3.4.x, so client IP is 1.2.3.4
        self.assertEqual(check_proxy(path_to_client,
                                     "192.168.1.1>10.0.0.0/24>2.3.4.0/24, 192.168.1.1>10.0.0.0/24>1.2.0.0/16"),
                         IPAddress("1.2.3.4"))
        # the order of proxy definitions is irrelevant
        self.assertEqual(check_proxy(path_to_client,
                                     "192.168.1.1>10.0.0.0/24>1.2.0.0/16, 192.168.1.1>10.0.0.0/24>2.3.4.0/24"),
                         IPAddress("1.2.3.4"))
        # proxy2 may map anywhere, and the next proxy may also map anywhere,
        # so we end up with 1.2.3.4
        self.assertEqual(check_proxy(path_to_client,
                                     "192.168.1.1>0.0.0.0/0>0.0.0.0/0"),
                         IPAddress("1.2.3.4"))
        # but if the next proxy may only map to 2.x.x.x, the proxy path does not match and we end up with proxy2.
        self.assertEqual(check_proxy(path_to_client,
                                     "192.168.1.1>0.0.0.0/0>2.0.0.0/8"),
                         IPAddress("192.168.1.1"))

        # another example
        path_to_client = list(map(IPAddress, ["10.1.1.1", "10.2.3.4", "192.168.1.1"]))
        self.assertEqual(check_proxy(path_to_client,
                                     "10.1.1.1/32>10.2.3.0/24>192.168.0.0/16"),
                         IPAddress("192.168.1.1"))
        self.assertEqual(check_proxy(path_to_client,
                                     "10.1.1.1/32>192.168.0.0/16"),
                         IPAddress("10.1.1.1"))
        self.assertEqual(check_proxy(path_to_client,
                                     "10.1.1.1/32>10.2.3.0/24>192.168.3.0/24"),
                         IPAddress("10.1.1.1"))

    def test_05_reduce_realms(self):
        realms = {'defrealm': {'default': False,
                               'option': '',
                               'resolver': [
                                   {'priority': None,
                                    'type': 'passwdresolver',
                                    'name': 'deflocal'}]},
                  'localsql': {'default': True,
                               'option': '',
                               'resolver': [
                                   {'priority': None,
                                    'type': 'sqlresolver',
                                    'name': 'localusers2'}]}}
        # The policy dictionary contains much more entries, but for us only
        # the realm is relevant
        policies = [{'realm': []}]
        r = reduce_realms(realms, policies)
        self.assertTrue("defrealm" in r)
        self.assertTrue("localsql" in r)

        policies = [{'realm': []},
                    {'realm': ["defrealm"]}]
        r = reduce_realms(realms, policies)
        self.assertTrue("defrealm" in r)
        self.assertTrue("localsql" in r)

        policies = [{'realm': ["defrealm"]},
                    {'realm': []}]
        r = reduce_realms(realms, policies)
        self.assertTrue("defrealm" in r)
        self.assertTrue("localsql" in r)

        policies = [{'realm': ["localsql"]},
                    {'realm': ["defrealm"]}]
        r = reduce_realms(realms, policies)
        self.assertTrue("defrealm" in r, r)
        self.assertTrue("localsql" in r, r)

        policies = [{'realm': ["localsql"]},
                    {'realm': ["localsql", "defrealm"]}]
        r = reduce_realms(realms, policies)
        self.assertTrue("defrealm" in r)
        self.assertTrue("localsql" in r)

        policies = [{'realm': ["realm1", "localsql", "realm2"]},
                    {'realm': ["localsql", "realm1"]}]
        r = reduce_realms(realms, policies)
        self.assertTrue("defrealm" not in r)
        self.assertTrue("localsql" in r)

        r = reduce_realms(realms, None)
        self.assertTrue("defrealm" in r)
        self.assertTrue("localsql" in r)

    def test_06_is_true(self):
        self.assertFalse(is_true(None))
        self.assertFalse(is_true(0))
        self.assertFalse(is_true("0"))
        self.assertFalse(is_true(False))
        self.assertFalse(is_true("false"))

        self.assertTrue(is_true(1))
        self.assertTrue(is_true("1"))
        self.assertTrue(is_true("True"))
        self.assertTrue(is_true("TRUE"))
        self.assertTrue(is_true(True))

    def test_07_parse_date(self):
        d = parse_date("+12m")
        self.assertTrue(datetime.now(tzlocal()) < d)

        d = parse_date(" +12m ")
        self.assertTrue(datetime.now(tzlocal()) < d)

        d = parse_date(" +12H ")
        self.assertTrue(datetime.now(tzlocal()) + timedelta(hours=11) < d)

        d = parse_date(" +12d ")
        self.assertTrue(datetime.now(tzlocal()) + timedelta(days=11) < d)

        d = parse_date("+5")
        self.assertTrue(datetime.now(tzlocal()) >= d)

        d = parse_date("")
        self.assertTrue(datetime.now(tzlocal()) >= d)

        self.assertIsNone(parse_date("-2g"))

        d = parse_date("2016/12/23")
        self.assertEqual(d, datetime(2016, 12, 23))

        d = parse_date("23.12.2016")
        self.assertEqual(d, datetime(2016, 12, 23))

        d = parse_date("2016/12/23 9:30pm")
        self.assertEqual(d, datetime(2016, 12, 23, hour=21, minute=30))

        d = parse_date("2016/12/23 10:30am")
        self.assertEqual(d, datetime(2016, 12, 23, hour=10, minute=30))

        d = parse_date("23.12.2016 21:30")
        self.assertEqual(d, datetime(2016, 12, 23, hour=21, minute=30))

        d = parse_date("23.12.2016 6:30")
        self.assertEqual(d, datetime(2016, 12, 23, hour=6, minute=30))

        d = parse_date("23.12.16")
        self.assertEqual(d, datetime(2016, 12, 23, 0, 0))

        d = parse_date("2017-04-27T12:00+0200")
        self.assertEqual(d, datetime(2017, 4, 27, 12, 0,
                                     tzinfo=tzoffset(None, 7200)))

        d = parse_date("2016/04/03")
        # April 3rd
        self.assertEqual(d, datetime(2016, 4, 3, 0, 0))

        d = parse_date("03.04.2016")
        # April 3rd
        self.assertEqual(d, datetime(2016, 4, 3, 0, 0))
        self.assertEqual(parse_date('03/04/2016'), datetime(2016, 4, 3, 0, 0))
        self.assertEqual(parse_date('01/01/20'), datetime(2020, 1, 1, 0, 0))
        self.assertEqual(parse_date('01/15/20'), datetime(2020, 1, 15, 0, 0))
        self.assertEqual(parse_date('15/01/20'), datetime(2020, 1, 15, 0, 0))

        # Non matching date returns None
        self.assertEqual(parse_date("7 Januar 17"), None)
        self.assertIsNone(parse_date('15/15'))

    def test_08_compare_condition(self):
        self.assertTrue(compare_condition("100", 100))
        self.assertTrue(compare_condition("=100", 100))
        self.assertTrue(compare_condition(" = 100 ", 100))

        self.assertFalse(compare_condition("100 ", 99))

        self.assertTrue(compare_condition(">100", 101))
        self.assertFalse(compare_condition(">100", 100))
        self.assertFalse(compare_condition(">100", 1))

        self.assertTrue(compare_condition("<100", 10))
        self.assertTrue(compare_condition("  <100", 10))
        self.assertFalse(compare_condition("<100", 1000))
        self.assertFalse(compare_condition("<100", 100))

        # There are invalid conditions, which should not raise an exception
        # An empty condition will result in False
        self.assertFalse(compare_condition("", 100))
        # An invalid condition, which misses a compare-value, will result in false
        self.assertFalse(compare_condition(">", 100))

        # Test new comparators
        self.assertTrue(compare_condition('>=100', 100))
        self.assertTrue(compare_condition('=> 100', 200))
        self.assertFalse(compare_condition('>= 100', 99))

        self.assertTrue(compare_condition('<=100', 100))
        self.assertTrue(compare_condition('=< 100', 99))
        self.assertFalse(compare_condition('<= 100', 101))

        self.assertTrue(compare_condition('!=100', 99))
        self.assertFalse(compare_condition('!= 100', 100))

        self.assertTrue(compare_condition('==100', 100))
        self.assertFalse(compare_condition('== 100', 99))

    def test_09_get_data_from_params(self):
        config_description = {
            "local": {
                'cakey': 'string',
                'cacert': 'string',
                'openssl.cnf': 'string',
                'WorkingDir': 'string',
                'CSRDir': 'sting',
                'CertificateDir': 'string',
                'CRLDir': 'string',
                'CRL_Validity_Period': 'int',
                'CRL_Overlap_Period': 'int'}}
        params = {"type": "local", "cakey": "key", "cacert": "cert",
                  "bindpw": "secret", "type.bindpw": "password"}
        data, types, desc = get_data_from_params(params,
                                                 ["caconnector", "type"],
                                                 config_description,
                                                 "CA connector",
                                                 "local")
        self.assertEqual(data.get("cakey"), "key")
        self.assertEqual(data.get("bindpw"), "secret")
        self.assertEqual(types.get("bindpw"), "password")

    def test_10_parse_legacy_time(self):
        s = parse_legacy_time("01/04/17 10:00")
        self.assertTrue(s.startswith("2017-04-01T10:00"))

        s = parse_legacy_time("30/04/17 10:00")
        self.assertTrue(s.startswith("2017-04-30T10:00"))

        s = parse_legacy_time("2017-04-01T10:00+0200")
        self.assertEqual(s, "2017-04-01T10:00+0200")

    def test_11_int_to_hex(self):
        h = int_to_hex(32)
        self.assertEqual(h, "20")

        h = int_to_hex(1)
        self.assertEqual(h, "01")

        h = int_to_hex(10)
        self.assertEqual(h, "0A")

        h = int_to_hex(256)
        self.assertEqual(h, "0100")

        h = int_to_hex(4096)
        self.assertEqual(h, "1000")

        h = int_to_hex(65536)
        self.assertEqual(h, "010000")

    def test_12_compare_value_value(self):
        self.assertTrue(compare_value_value("1000", ">", "999"))
        self.assertTrue(compare_value_value("ABD", ">", "ABC"))
        self.assertTrue(compare_value_value(1000, "==", "1000"))
        self.assertTrue(compare_value_value("99", "<", "1000"))

        self.assertTrue(compare_value_value(100, '>', '10'))
        self.assertTrue(compare_value_value(100, '>=', '10'))
        self.assertTrue(compare_value_value("100", '=>', 10))
        self.assertTrue(compare_value_value(100, '>=', '100'))
        self.assertFalse(compare_value_value(100, '>=', '101'))

        self.assertTrue(compare_value_value('ABC', '=', 'ABC'))
        self.assertTrue(compare_value_value('ABC', '!=', 'ABD'))

        self.assertTrue(compare_value_value(10, '<', '100'))
        self.assertTrue(compare_value_value(10, '<=', '100'))
        self.assertTrue(compare_value_value("10", '=<', 100))
        self.assertTrue(compare_value_value(10, '<=', '10'))
        self.assertFalse(compare_value_value(10, '<=', '9'))

        # compare dates
        self.assertTrue(compare_value_value(
            datetime.now(tzlocal()).strftime(DATE_FORMAT), ">",
            "2017-01-01T10:00+0200"))
        self.assertFalse(compare_value_value(
            datetime.now(tzlocal()).strftime(DATE_FORMAT), "<",
            "2017-01-01T10:00+0200"))
        # The timestamp in 10 hours is bigger than the current time
        self.assertTrue(compare_value_value(
            (datetime.now(tzlocal()) + timedelta(hours=10)).strftime(DATE_FORMAT),
            ">", datetime.now(tzlocal()).strftime(DATE_FORMAT)))

        self.assertTrue(compare_value_value('+3h', '>', ''))
        self.assertFalse(compare_value_value('2017/04/20 11:30+0200', '>',
                                             datetime.now(tzlocal()).strftime(DATE_FORMAT)))

        self.assertTrue(compare_value_value('2020-01-15T00:00', '==',
                                            datetime(2020, 1, 15).strftime(DATE_FORMAT)))
        # unexpected result: The date string can not be parsed since dateutil.parser
        # does not understand locale dates. So the strings themselves are compared
        # since parse_date() returns 'None'
        self.assertTrue(compare_value_value('16. März 2020', '<',
                                            datetime(2020, 3, 15).strftime(DATE_FORMAT)))

        # check for unknown comparator
        self.assertRaises(Exception, compare_value_value, 5, '~=', 5)

    def test_13_parse_time_offset_from_now(self):
        td = parse_timedelta("+5s")
        self.assertEqual(td, timedelta(seconds=5))
        td = parse_timedelta("-12m")
        self.assertEqual(td, timedelta(minutes=-12))
        td = parse_timedelta("+123h")
        self.assertEqual(td, timedelta(hours=123))
        td = parse_timedelta("+2d")
        self.assertEqual(td, timedelta(days=2))

        # It is allowed to start without a +/- which would mean a +
        td = parse_timedelta("12d")
        self.assertEqual(td, timedelta(days=12))

        # Does not contains numbers
        self.assertRaises(Exception, parse_timedelta, "+twod")

        s, td = parse_time_offset_from_now("Hello {now}+5d with 5 days.")
        self.assertEqual(s, "Hello {now} with 5 days.")
        self.assertEqual(td, timedelta(days=5))

        s, td = parse_time_offset_from_now("Hello {current_time}+5m!")
        self.assertEqual(s, "Hello {current_time}!")
        self.assertEqual(td, timedelta(minutes=5))

        s, td = parse_time_offset_from_now("Hello {current_time}-3habc")
        self.assertEqual(s, "Hello {current_time}abc")
        self.assertEqual(td, timedelta(hours=-3))

    def test_14_convert_timestamp_to_utc(self):
        d = datetime.now(tz=gettz('America/New York'))
        d_utc = convert_timestamp_to_utc(d)
        self.assertGreater(d_utc, d.replace(tzinfo=None))

    def test_16_parse_int(self):
        r = parse_int("xxx", 12)
        self.assertEqual(r, 12)
        r = parse_int("ABC", 11)
        self.assertEqual(r, 2748)
        r = parse_int("ABCX", 11)
        self.assertEqual(r, 11)
        r = parse_int(123)
        self.assertEqual(r, 123)
        r = parse_int(0x12)
        self.assertEqual(r, 18)
        r = parse_int("0x12")
        self.assertEqual(r, 18)
        r = parse_int("123")
        self.assertEqual(r, 123)

    def test_17_convert_column_to_unicode(self):
        self.assertEqual(convert_column_to_unicode(None), None)
        self.assertEqual(convert_column_to_unicode(True), "True")
        self.assertEqual(convert_column_to_unicode(False), "False")
        self.assertEqual(convert_column_to_unicode(b"yes"), "yes")
        self.assertEqual(convert_column_to_unicode("yes"), "yes")

    def test_18_censor_connect_string(self):
        self.assertEqual(censor_connect_string("sqlite:////home/foo/privacyidea/privacyidea/data.sqlite"),
                         "sqlite:////home/foo/privacyidea/privacyidea/data.sqlite")
        self.assertEqual(censor_connect_string("mysql://pi@localhost/pi"),
                         "mysql://pi@localhost/pi")
        self.assertEqual(censor_connect_string("mysql://pi:kW44sqqWtGYX@localhost/pi"),
                         "mysql://pi:***@localhost/pi")
        self.assertEqual(censor_connect_string("psql+odbc://pi@localhost/pi"),
                         "psql+odbc://pi@localhost/pi")
        self.assertEqual(censor_connect_string("psql+odbc://pi:MySecretPassword123466$@localhost/pi"),
                         "psql+odbc://pi:***@localhost/pi")
        self.assertEqual(censor_connect_string("mysql://pi:kW44s%40%40qqWtGYX@localhost/pi"),
                         "mysql://pi:***@localhost/pi")
        self.assertEqual(censor_connect_string("mysql://knöbel:föö@localhost/pi"),
                         "mysql://knöbel:***@localhost/pi")
        self.assertEqual(censor_connect_string(
            "oracle+cx_oracle://pi:MySecretPassword1234@localhost:1521/?service_name=my_database"),
            "oracle+cx_oracle://pi:***@localhost:1521/?service_name=my_database")

    def test_19_truncate_comma_list(self):
        r = truncate_comma_list("123456,234567,345678", 19)
        self.assertEqual(len(r), 19)
        self.assertEqual(r, "1234+,234567,345678")

        r = truncate_comma_list("123456,234567,345678", 18)
        self.assertEqual(len(r), 18)
        self.assertEqual(r, "1234+,2345+,345678")

        r = truncate_comma_list("123456,234567,345678", 16)
        self.assertEqual(len(r), 16)
        self.assertEqual(r, "123+,2345+,3456+")

        # There are more entries than the max_len. We will not be able
        # to shorten all entries, so we simply take the beginning of the string.
        r = truncate_comma_list("12,234567,3456,989,123,234,234", 4)
        self.assertEqual(len(r), 4)
        self.assertEqual(r, "12,+")

    def test_20_pin_policy(self):
        # Unspecified character specifier
        self.assertRaises(PolicyError, check_pin_contents, "1234", "+o")

        r, c = check_pin_contents("1234", "n")
        self.assertTrue(r)

        r, c = check_pin_contents(r"[[[", "n")
        self.assertFalse(r)

        r, c = check_pin_contents(r"[[[", "c")
        self.assertFalse(r)

        r, c = check_pin_contents(r"[[[", "s")
        self.assertTrue(r)

        # check the validation of a generated password with square brackets
        password = generate_password(size=3, requirements=['[', '[', '['])
        r, c = check_pin_contents(password, "s")
        self.assertTrue(r)

        r, c = check_pin_contents("abc", "nc")
        self.assertFalse(r)
        self.assertEqual("Missing character in PIN: {}".format(CHARLIST_CONTENTPOLICY['n']), c)

        r, c = check_pin_contents("123", "nc")
        self.assertFalse(r)
        self.assertEqual(r"Missing character in PIN: {}".format(CHARLIST_CONTENTPOLICY['c']), c)

        r, c = check_pin_contents("123", "ncs")
        self.assertFalse(r)
        self.assertTrue(r"Missing character in PIN: {}".format(CHARLIST_CONTENTPOLICY['c'] in c), c)
        self.assertTrue(r"Missing character in PIN: {}".format(CHARLIST_CONTENTPOLICY['s'] in c), c)

        r, c = check_pin_contents("1234", "")
        self.assertFalse(r)
        self.assertEqual(c, "No policy given.")

        # check for either number or character
        r, c = check_pin_contents("1234", "+cn")
        self.assertTrue(r)

        r, c = check_pin_contents("1234xxxx", "+cn")
        self.assertTrue(r)

        r, c = check_pin_contents("xxxx", "+cn")
        self.assertTrue(r)
        self.assertTrue(check_pin_contents("test1234", "+cn")[0])
        self.assertTrue(check_pin_contents("test12$$", "+cn")[0])
        self.assertTrue(check_pin_contents("test12", "+cn")[0])
        self.assertTrue(check_pin_contents("1234", "+cn")[0])

        r, c = check_pin_contents("@@@@", "+cn")
        self.assertFalse(r)
        self.assertEqual(c, "Missing character in PIN: {}{}".format(CHARLIST_CONTENTPOLICY['c'],
                                                                    CHARLIST_CONTENTPOLICY['n']))

        # check for exclusion
        # No special character
        r, c = check_pin_contents("1234", "-s")
        self.assertTrue(r)
        r, c = check_pin_contents("1234aaaa", "-s")
        self.assertTrue(r)
        r, c = check_pin_contents("1234aaaa//", "-s")
        self.assertFalse(r)

        # A pin that falsely contains a number
        r, c = check_pin_contents("1234aaa", "-sn")
        self.assertFalse(r)
        self.assertEqual(c, "Not allowed character in PIN!")
        r, c = check_pin_contents("///aaa", "-sn")
        self.assertFalse(r)
        # A pin without a number and without a special
        r, c = check_pin_contents("xxxx", "-sn")
        self.assertTrue(r)

        r, c = check_pin_contents("1234@@@@", "-c")
        self.assertTrue(r)

        # A pin with only digits allowed
        r, c = check_pin_contents("1234", "-cs")
        self.assertTrue(r)
        r, c = check_pin_contents("a1234", "-cs")
        self.assertFalse(r)

        # A pin with only a specified list of chars
        r, c = check_pin_contents("1234111", "[1234]")
        self.assertTrue(r)
        r, c = check_pin_contents("12345", "[1234]")
        self.assertFalse(r)

    def test_21_get_module_class(self):
        r = get_module_class("privacyidea.lib.auditmodules.sqlaudit", "Audit", "log")
        from privacyidea.lib.auditmodules.sqlaudit import Audit
        self.assertEqual(r, Audit)

        # Fails to return the class, if the method does not exist
        self.assertRaises(NameError, get_module_class, "privacyidea.lib.auditmodules.sqlaudit", "Audit",
                          "this_method_does_not_exist")

        # Fails if the class does not exist
        with self.assertRaises(ImportError):
            get_module_class("privacyidea.lib.auditmodules.sqlaudit", "DoesNotExist")

        # Fails if the package does not exist
        with self.assertRaises(ImportError):
            get_module_class("privacyidea.lib.auditmodules.doesnotexist", "Aduit")

    def test_22_decodebase32check(self):
        real_client_componet = "TIXQW4ydvn2aos4cj6ta"
        real_payload = "03ab74074b824fa6"

        payload1 = decode_base32check(real_client_componet)
        self.assertEqual(payload1, real_payload)

        # change the client component in the last character!
        client_component = "TIXQW4ydvn2aos4cj6tb"
        payload2 = decode_base32check(client_component)
        # Although the last character of the client component was changed,
        # the payload is still the same.
        self.assertEqual(payload2, real_payload)

        # change the client component in between
        client_component = "TIXQW4ydvn2aos4cj6ba"
        self.assertRaises(Exception, decode_base32check, client_component)

    def test_23_get_client_ip(self):
        class RequestMock():
            blueprint = None
            remote_addr = None
            all_data = {}
            access_route = []

        r = RequestMock()
        r.blueprint = "token_blueprint"
        # The real client
        direct_client = "10.0.0.1"
        r.remote_addr = direct_client
        # The client parameter
        client_parameter = "192.168.2.1"
        r.all_data = {"client": client_parameter}
        # The X-Forwarded-For
        client_proxy = "172.16.1.2"
        r.access_route = [client_proxy]

        # Setup:
        # 192.168.2.1 ---------> 172.16.1.2 -------> 10.0.0.1 --------> privacyIDEA
        # client_parameter       client_proxy        direct_client

        ip = get_client_ip(r, "")
        self.assertEqual(ip, direct_client)

        # If there is a proxy_setting, the X-Forwarded-For will
        # work, but not the client_parameter
        ip = get_client_ip(r, "10.0.0.1")
        self.assertEqual(ip, client_proxy)

        # If the request is a validate request:
        r.blueprint = "validate_blueprint"
        # ... the direct client may map anywhere, but as we also have a X-Forwarded-For header,
        # the header takes precedence!
        ip = get_client_ip(r, "10.0.0.1")
        self.assertEqual(ip, client_proxy)
        # ... if we now also allow the client_proxy to rewrite IPs, the client parameter is respected
        ip = get_client_ip(r, "10.0.0.1>172.16.1.2>0.0.0.0/0")
        self.assertEqual(ip, client_parameter)
        # ... even if we have multiple proxy settings
        ip = get_client_ip(r, "10.0.0.1>198.168.1.3, 10.0.0.1>172.16.1.2>1.2.3.4,   10.0.0.1>172.16.1.2>0.0.0.0/0")
        self.assertEqual(ip, client_parameter)
        # Check situation if there is no X-Forwarded-For header, but a client parameter:
        r.access_route = [direct_client]
        ip = get_client_ip(r, "10.0.0.1")
        self.assertEqual(ip, client_parameter)
        # The client parameter is not respected for the token endpoints
        r.blueprint = "token_blueprint"
        ip = get_client_ip(r, "10.0.0.1")
        self.assertEqual(ip, direct_client)

    def test_24_sanity_name_check(self):
        self.assertTrue(sanity_name_check('Hello_World'))
        with self.assertRaisesRegex(Exception, "non conformant characters in the name"):
            sanity_name_check('Hello World!')
        self.assertTrue(sanity_name_check('Hello World', name_exp='^[A-Za-z\\ ]+$'))
        with self.assertRaisesRegex(Exception, "non conformant characters in the name"):
            sanity_name_check('Hello_World', name_exp='^[A-Za-z]+$')

    def test_25_encodings(self):
        u = 'Hello Wörld'
        b = b'Hello World'
        self.assertEqual(to_utf8(None), None)
        self.assertEqual(to_utf8(u), u.encode('utf8'))
        self.assertEqual(to_utf8(b), b)

        self.assertEqual(to_unicode(u), u)
        self.assertEqual(to_unicode(b), b.decode('utf8'))
        self.assertEqual(to_unicode(None), None)
        self.assertEqual(to_unicode(10), 10)

        self.assertEqual(to_bytes(u), u.encode('utf8'))
        self.assertEqual(to_bytes(b), b)
        self.assertEqual(to_bytes(10), 10)

        self.assertEqual(to_byte_string(u), u.encode('utf8'))
        self.assertEqual(to_byte_string(b), b)
        self.assertEqual(to_byte_string(10), b'10')

    def test_26_conversions(self):
        self.assertEqual(hexlify_and_unicode('Hallo'), '48616c6c6f')
        self.assertEqual(hexlify_and_unicode(b'Hallo'), '48616c6c6f')
        self.assertEqual(hexlify_and_unicode(b'\x00\x01\x02\xab'), '000102ab')

        self.assertEqual(b32encode_and_unicode('Hallo'), 'JBQWY3DP')
        self.assertEqual(b32encode_and_unicode(b'Hallo'), 'JBQWY3DP')
        self.assertEqual(b32encode_and_unicode(b'\x00\x01\x02\xab'), 'AAAQFKY=')

        self.assertEqual(b64encode_and_unicode('Hallo'), 'SGFsbG8=')
        self.assertEqual(b64encode_and_unicode(b'Hallo'), 'SGFsbG8=')
        self.assertEqual(b64encode_and_unicode(b'\x00\x01\x02\xab'), 'AAECqw==')

        self.assertEqual(urlsafe_b64encode_and_unicode('Hallo'), 'SGFsbG8=')
        self.assertEqual(urlsafe_b64encode_and_unicode(b'Hallo'), 'SGFsbG8=')
        self.assertEqual(urlsafe_b64encode_and_unicode(b'\x00\x01\x02\xab'), 'AAECqw==')
        self.assertEqual(urlsafe_b64encode_and_unicode(b'\xfa\xfb\xfc\xfd\xfe\xff'),
                         '-vv8_f7_')

    def test_27_images(self):
        hallo_qr_png = "iVBORw0KGgoAAAANSUhEUgAAASIAAAEiAQAAAAB1xeIbAAABC0lEQV" \
                       "R42u2aQQ6EIBAEJ7sP8El8nSftA0xYGBiM8eIe6E1McTCofSpnmka1" \
                       "cmNkQ4UKFSpUj1DZGO86//ghriRXvezOQPWbarB3xBP7mM0bsF/Kvh" \
                       "V6asRzFL+3AezV7H0GezV7s2036v4/fp8r+94B+L2G/cw5vfgTOUfG" \
                       "/hgV9n5Jp7Bfyr4nSzf92QGwl6211epLZMwSCy6es7zu83aC3di7+8" \
                       "BelDFtRpzeAVs4P+zXrrWVc26nx742kTHVOad3QCn4vTzfz1RPztHv" \
                       "a3u8DAuCve59ToR8fwrsa6Xs4wEM96Hulez9PeaM+7CX+n0P+acFF/" \
                       "aSnBOfcY26l+d7/i1AhQoVqmeqvi4sW6dMYAvIAAAAAElFTkSuQmCC"
        self.assertEqual(create_img('Hallo'), 'data:image/png;base64,{0!s}'.format(hallo_qr_png))

    def test_28_yubikey_utils(self):
        self.assertEqual(modhex_encode(b'\x47'), 'fi')
        self.assertEqual(modhex_encode(b'\xba\xad\xf0\x0d'), 'nlltvcct')
        self.assertEqual(modhex_encode(binascii.unhexlify('0123456789abcdef')),
                         'cbdefghijklnrtuv')
        self.assertEqual(modhex_encode('Hallo'), 'fjhbhrhrhv')
        # and the other way around
        self.assertEqual(modhex_decode('fi'), b'\x47')
        self.assertEqual(modhex_decode('nlltvcct'), b'\xba\xad\xf0\x0d')
        self.assertEqual(modhex_decode('cbdefghijklnrtuv'),
                         binascii.unhexlify('0123456789abcdef'))
        self.assertEqual(modhex_decode('fjhbhrhrhv'), b'Hallo')
        # fail with invalid modhex
        self.assertRaises((binascii.Error, TypeError), modhex_decode, 'nlltvcc')

        # now test the crc function
        self.assertEqual(checksum(b'\x01\x02\x03\x04'), 0xc66e)
        self.assertEqual(checksum(b'\x01\x02\x03\x04\x919'), 0xf0b8)

    def test_29_check_ip(self):
        found, excluded = check_ip_in_policy("10.0.1.2", ["10.0.1.0/24", "1.1.1.1"])
        self.assertTrue(found)
        self.assertFalse(excluded)

        found, excluded = check_ip_in_policy("10.0.1.2", ["10.0.1.0/24", "!10.0.1.2"])
        self.assertTrue(excluded)
        self.assertTrue(found)

        # run a test for empty condition
        found, excluded = check_ip_in_policy("10.0.1.2", ["10.0.1.0/24", "!10.0.1.2", '', None])
        self.assertTrue(excluded)
        self.assertTrue(found)

    def test_30_split_pin_pass(self):
        pin, otp = split_pin_pass("test1234", 4, True)
        self.assertEqual(pin, "test")
        self.assertEqual(otp, "1234")
        pin, otp = split_pin_pass("12345678hallo", 8, False)
        self.assertEqual(pin, "hallo")
        self.assertEqual(otp, "12345678")

    def test_31_create_tag_dict(self):
        class UserAgentMock():
            string = "<b>hello world</b>"
            browser = "browser"

        class RequestMock():
            user_agent = UserAgentMock()
            path = "/validate/check"
            url_root = ""

        recipient = {"givenname": "<b>Sömeone</b>"}
        dict1 = create_tag_dict(request=RequestMock(), recipient=recipient)
        self.assertEqual(dict1["ua_string"], "<b>hello world</b>")
        self.assertEqual(dict1["action"], "/validate/check")
        self.assertEqual(dict1["recipient_givenname"], "<b>Sömeone</b>")
        dict2 = create_tag_dict(request=RequestMock(), recipient=recipient, escape_html=True)
        self.assertEqual(dict2["ua_string"], "&lt;b&gt;hello world&lt;/b&gt;")
        self.assertEqual(dict2["action"], "/validate/check")
        self.assertEqual(dict2["recipient_givenname"], "&lt;b&gt;Sömeone&lt;/b&gt;")

    def test_32_allowed_serial_numbers(self):
        self.assertTrue(check_serial_valid("TOTP12345"))
        # Blank is not allowed
        self.assertRaises(Exception, check_serial_valid, "TOTP 12345")

        # Minus and underscore is allowed
        self.assertTrue(check_serial_valid("spass-123"))
        self.assertTrue(check_serial_valid("spass_123"))
        # Slash and backslash is not allowed
        self.assertRaises(Exception, check_serial_valid, "spass/123")
        self.assertRaises(Exception, check_serial_valid, "spass\\123")

        # an empty serial is not allowed
        self.assertRaises(Exception, check_serial_valid, "")

    def test_33_determine_logged_in_user(self):
        (role, user, realm, adminuser, adminrealm) = determine_logged_in_userparams({"role": "user",
                                                                                     "username": "hans",
                                                                                     "realm": "realm1"}, {})

        self.assertEqual(role, "user")
        self.assertEqual(user, "hans")
        self.assertEqual(realm, "realm1")
        self.assertEqual(adminuser, None)
        self.assertEqual(adminrealm, None)

        (role, user, realm, adminuser, adminrealm) = determine_logged_in_userparams({"role": "admin",
                                                                                     "username": "hans",
                                                                                     "realm": "realm1"},
                                                                                    {"user": "peter",
                                                                                     "realm": "domain"})

        self.assertEqual(role, "admin")
        self.assertEqual(user, "peter")
        self.assertEqual(realm, "domain")
        self.assertEqual(adminuser, "hans")
        self.assertEqual(adminrealm, "realm1")

        self.assertRaises(PolicyError, determine_logged_in_userparams,
                          {"role": "marshal",
                           "username": "Wyatt Earp",
                           "realm": "Wild West"},
                          {"user": "Dave Rudabaugh",
                           "realm": "Dodge City"})

    def test_34_compare_generic_condition(self):
        def mock_attribute(key):
            attr = {"a": "10",
                    "b": "100",
                    "c": "1000"}
            return attr.get(key)

        self.assertTrue(compare_generic_condition("a<100",
                                                  mock_attribute,
                                                  "Error {0!s}"))

        self.assertTrue(compare_generic_condition("a <100",
                                                  mock_attribute,
                                                  "Error {0!s}"))

        self.assertTrue(compare_generic_condition("b==100",
                                                  mock_attribute,
                                                  "Error {0!s}"))

        # Wrong condition
        self.assertFalse(compare_generic_condition("a== 100",
                                                   mock_attribute,
                                                   "Error {0!s}"))

        # Wrong condition
        self.assertFalse(compare_generic_condition("b>100",
                                                   mock_attribute,
                                                   "Error {0!s}"))

        # Wrong condition
        self.assertFalse(compare_generic_condition("c < 500",
                                                   mock_attribute,
                                                   "Error {0!s}"))

        # Wrong condition
        self.assertFalse(compare_generic_condition("c <500",
                                                   mock_attribute,
                                                   "Error {0!s}"))

        # Wrong condition: key does not exist
        self.assertFalse(compare_generic_condition("d==1",
                                                   mock_attribute,
                                                   "Error {0!s}"))

        # Wrong condition: key does not exist
        self.assertFalse(compare_generic_condition("d<1",
                                                   mock_attribute,
                                                   "Error {0!s}"))

        # Wrong condition: key does not exist
        self.assertFalse(compare_generic_condition("d>1",
                                                   mock_attribute,
                                                   "Error {0!s}"))

        # Wrong entry, that is not processed
        self.assertRaises(Exception, compare_generic_condition,
                          "c 500", mock_attribute, "Error {0!s}")

        # Wrong entry, that cannot be processed
        self.assertRaises(Exception, compare_generic_condition,
                          "b!~100", mock_attribute, "Error {0!s}")

    def test_34_to_list(self):
        # Simple string
        self.assertEqual(["Hallo"], to_list("Hallo"))
        # check for a list
        r = to_list(["Hallo", "Du", "da"])
        self.assertIsInstance(r, list)
        self.assertEqual(3, len(r))
        # Check for a set
        r = to_list({"Hallo", "Du", "da"})
        self.assertIsInstance(r, list)
        self.assertEqual(3, len(r))

    def test_35_parse_string_to_dict(self):
        # We can have as many whitespaces
        d = parse_string_to_dict(" :key1: v1 v2  v3:key2: v4 v5 ")
        self.assertIn("key1", d)
        self.assertIn("key2", d)
        self.assertEqual(d.get("key1"), ["v1", "v2", "v3"])
        self.assertEqual(d.get("key2"), ["v4", "v5"])

        d = parse_string_to_dict(" :key1: v1 v2  *:*: v4 v5 ")
        self.assertIn("key1", d)
        self.assertIn("*", d)
        self.assertEqual(d.get("key1"), ["v1", "v2", "*"])
        self.assertEqual(d.get("*"), ["v4", "v5"])

        # we can have no whitespaces
        d = parse_string_to_dict(":key1:v1 v2 v3 :key2:v4 v5")
        self.assertIn("key1", d)
        self.assertIn("key2", d)
        self.assertEqual(d.get("key1"), ["v1", "v2", "v3"])
        self.assertEqual(d.get("key2"), ["v4", "v5"])

        # side effect: we can skip the leading colon
        d = parse_string_to_dict("key1:v1 v2 v3 :key2:v4 v5")
        self.assertIn("key1", d)
        self.assertIn("key2", d)
        self.assertEqual(d.get("key1"), ["v1", "v2", "v3"])
        self.assertEqual(d.get("key2"), ["v4", "v5"])

        # errors
        d = parse_string_to_dict("key1:v1 v2 v3 :key2::v4 v5")
        self.assertIn("key1", d)
        self.assertIn("key2", d)
        self.assertEqual(d.get("key1"), ["v1", "v2", "v3"])
        self.assertEqual(d.get("key2"), ["v4", "v5"])

        d = parse_string_to_dict(":key1:v1 v2 v3 :key2: ::key3: v5")
        self.assertIn("key1", d)
        self.assertIn("key2", d)
        self.assertIn("key3", d)
        self.assertEqual(d.get("key1"), ["v1", "v2", "v3"])
        self.assertEqual(d.get("key2"), [])
        self.assertEqual(d.get("key3"), ["v5"])

    def test_36_imagefile_to_dataimage(self):
        file = "./tests/testdata/FIDO-U2F-Security-Key-444x444.png"
        dataimage = convert_imagefile_to_dataimage(file)
        self.assertTrue(dataimage.startswith("data:image/png;base64,iVBOR"))

        # File not found returns an empty datatime string
        file = "./tests/testdata/FIDO-U2F-Security-Key-444x444.XXX"
        dataimage = convert_imagefile_to_dataimage(file)
        self.assertEqual("", dataimage)

        # Try to read a grazy file with an unknown mime-type
        file = "./tests/testdata/logging.yml"
        dataimage = convert_imagefile_to_dataimage(file)
        self.assertEqual("", dataimage)

    def test_37_useragent_split(self):
        user_agents = [
            ("privacyidea-cp", "privacyidea-cp", None, None),
            ("privacyidea-cp/2.0", "privacyidea-cp", "2.0", None),
            ("Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:47.0) Gecko/20100101 "
             "Firefox/47.0", "Mozilla", "5.0",
             "(Windows NT 6.1; Win64; x64; rv:47.0) Gecko/20100101 Firefox/47.0"),
            ("privacyIDEA-Keycloak", "privacyIDEA-Keycloak", None, None),
            ("privacyIDEA-Keycloak/20 java/8", "privacyIDEA-Keycloak", "20", "java/8"),
            ("PrivacyIDEA-ADFS", "PrivacyIDEA-ADFS", None, None),
            ("simpleSAMLphp", "simpleSAMLphp", None, None),
            ("simpleSAMLphp/1.2.3", "simpleSAMLphp", "1.2.3", None),
            ("ownCloud", "ownCloud", None, None),
            ("privacyIDEA-PAM", "privacyIDEA-PAM", None, None),
            ("privacyIDEA-Shibboleth", "privacyIDEA-Shibboleth", None, None),
            ("privacyIDEA-LDAP-Proxy", "privacyIDEA-LDAP-Proxy", None, None),
            ("privacyIDEA-LDAP-Proxy/10", "privacyIDEA-LDAP-Proxy", "10", None),
            # matching fails completely
            ("#test", "", None, None),
            # matching fails after first dot
            ("1.2.3", "1", None, None),
            # matching stops after missing version
            ("SomePlugin/ Some Comment", "SomePlugin", None, None),
            # no user-agent string given
            ("", "", None, None),
            (None, "", None, None)
        ]

        for val in user_agents:
            res = get_plugin_info_from_useragent(val[0])
            self.assertEqual(res, val[1:], res)


class UtilsTestCaseOverrideConfig(OverrideConfigTestCase):
    class Config(TestingConfig):
        OFFLINE_MACHINE_KEYS = ["Hostname", "myMachineIdentifier", "otherMachineIdentifier"]

    def test_01_get_computer_name(self):
        data = {
            "privacyidea-cp/1.1.1 Windows/Laptop-1": "Laptop-1",
            "privacyidea-cp/1.1.1 ComputerName/Laptop-2": "Laptop-2",
            "privacyidea-pam/2.2.2 Linux/Server-3": "Server-3",
            "privacyidea-pam/2.2.2 Hostname/Server-4": "Server-4",
            "privacyidea-mac/3.3.3 Mac/Server-5": "Server-5",
            "": None,
            None: None,
            "privacyidea-mac/3.3.3": None,
            "privacyidea-mac/3.3.3 myMachineIdentifier/Toaster": "Toaster",
            # This is also "Toaster" because the order of the keys is fixed, therefore "myMachineIdentifier" is checked
            # first and found
            "otherMachineIdentifier/Dishwasher myMachineIdentifier/Toaster": "Toaster"
        }

        for k, v in data.items():
            res = get_computer_name_from_user_agent(k)
            self.assertEqual(v, res)



#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2019-01-14   Cornelius Kölbel <cornelius.koelbel@netknights.it>
#               Allow different SMS gateways via "sms.identifier"
#  2018-10-31   Cornelius Kölbel <cornelius.koelbel@netknights.it>
#               Let the client choose to get a HTTP 500 Error code if
#               SMS fails.
#  2018-02-16   Cornelius Kölbel <cornelius.koelbel@netknights.it>
#               Allow to use a dynamic_phone
#  2016-06-20   Cornelius Kölbel <cornelius.koelbel@netkngihts.it>
#               Use sms.identifier, central SMS gateway definition, to send
#               the OTP value via SMS.
#  2015-05-24   Add more detailed description
#               Cornelius Kölbel <cornelius.koelbel@netknights.it>
#  2015-01-30   Adapt for migration to flask
#               Cornelius Kölbel <cornelius@privacyidea.org>
#
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""The SMS token sends an SMS containing an OTP via some kind of
gateway. The gateways can be an SMTP or HTTP gateway or the special sipgate
protocol.
The Gateways are defined in the SMSProvider Modules.

This code is tested in tests/test_lib_tokens_sms
"""

import datetime
import logging
import traceback
from json import loads

from privacyidea.api.lib.utils import getParam
from privacyidea.api.lib.utils import required, optional
from privacyidea.lib import _
from privacyidea.lib.config import get_from_config
from privacyidea.lib.crypto import safe_compare
from privacyidea.lib.decorators import check_token_locked
from privacyidea.lib.log import log_with
from privacyidea.lib.policy import (SCOPE, ACTION, GROUP, comma_escape_text,
                                    Match, get_action_values_from_options)
from privacyidea.lib.smsprovider.SMSProvider import (get_sms_provider_class,
                                                     create_sms_instance,
                                                     get_smsgateway)
from privacyidea.lib.tokenclass import CHALLENGE_SESSION, AUTHENTICATIONMODE
from privacyidea.lib.tokens.hotptoken import HotpTokenClass
from privacyidea.lib.utils import is_true, create_tag_dict
from privacyidea.models import Challenge

log = logging.getLogger(__name__)

keylen = {'sha1': 20,
          'sha256': 32,
          'sha512': 64}


class SMSACTION(object):
    SMSTEXT = "smstext"
    SMSAUTO = "smsautosend"
    GATEWAYS = "sms_gateways"


class SmsTokenClass(HotpTokenClass):
    """
    The SMS token sends an SMS containing an OTP via some kind of
    gateway. The gateways can be an SMTP or HTTP gateway or the special sipgate
    protocol. The Gateways are defined in the SMSProvider Modules.

    The SMS token is a challenge response token. I.e. the first request needs
    to contain the correct OTP PIN. If the OTP PIN is correct, the sending of
    the SMS is triggered. The second authentication must either contain the
    OTP PIN and the OTP value or the transaction_id and the OTP value.

      **Example 1st Authentication Request**:

        .. sourcecode:: http

           POST /validate/check HTTP/1.1
           Host: example.com
           Accept: application/json

           user=cornelius
           pass=otppin

      **Example 1st response**:

           .. sourcecode:: http

               HTTP/1.1 200 OK
               Content-Type: application/json

               {
                  "detail": {
                    "transaction_id": "xyz"
                  },
                  "id": 1,
                  "jsonrpc": "2.0",
                  "result": {
                    "status": true,
                    "value": false
                  },
                  "version": "privacyIDEA unknown"
                }

    After this, the SMS is triggered. When the SMS is received the second part
    of authentication looks like this:

      **Example 2nd Authentication Request**:

        .. sourcecode:: http

           POST /validate/check HTTP/1.1
           Host: example.com
           Accept: application/json

           user=cornelius
           transaction_id=xyz
           pass=otppin

      **Example 1st response**:

           .. sourcecode:: http

               HTTP/1.1 200 OK
               Content-Type: application/json

               {
                  "detail": {
                  },
                  "id": 1,
                  "jsonrpc": "2.0",
                  "result": {
                    "status": true,
                    "value": true
                  },
                  "version": "privacyIDEA unknown"
                }


    """
    mode = [AUTHENTICATIONMODE.CHALLENGE]

    def __init__(self, db_token):
        HotpTokenClass.__init__(self, db_token)
        self.set_type("sms")
        self.hKeyRequired = True

    @staticmethod
    def get_class_type():
        """
        return the generic token class identifier
        """
        return "sms"

    @staticmethod
    def get_class_prefix():
        return "PISM"

    @staticmethod
    def get_class_info(key=None, ret='all'):
        """
        returns all or a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined

        :return: subsection if key exists or user defined
        :rtype : s.o.
        """
        sms_gateways = [gw.identifier for gw in get_smsgateway()]
        res = {'type': 'sms',
               'title': _('SMS Token'),
               'description':
                   _('SMS: Send a One Time Password to the users mobile '
                     'phone.'),
               'user': ['enroll'],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin", "user"],
               'policy': {
                   SCOPE.AUTH: {
                       SMSACTION.SMSTEXT: {
                           'type': 'str',
                           'desc': _('The text that will be send via SMS for '
                                     'an SMS token. Use tags like {otp} and {serial} '
                                     'as parameters.')
                                   + " " + comma_escape_text
                       },
                       SMSACTION.SMSAUTO: {
                           'type': 'bool',
                           'desc': _('If set, a new SMS OTP will be sent '
                                     'after successful authentication with '
                                     'one SMS OTP.')},
                       ACTION.CHALLENGETEXT: {
                           'type': 'str',
                           'desc': _('Use an alternative challenge text for telling the '
                                     'user to enter the code from the SMS. You can also '
                                     'use tags for automated replacement. Check out the '
                                     'documentation for more details.')
                                   + " " + comma_escape_text
                       }
                   },
                   SCOPE.ADMIN: {
                       SMSACTION.GATEWAYS: {
                           'type': 'str',
                           'desc': "{0!s} ({1!s})".format(
                               _('Choose the gateways the administrator is allowed to set.'),
                               " ".join(sms_gateways))
                       }
                   },
                   SCOPE.USER: {
                       SMSACTION.GATEWAYS: {
                           'type': 'str',
                           'desc': "{0!s} ({1!s})".format(
                               _('Choose the gateways the user is allowed to set.'),
                               " ".join(sms_gateways))
                       }
                   },
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of SMS tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _(
                               "The user may only have this maximum number of active SMS tokens assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               },
               }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res

        return ret

    @log_with(log)
    def update(self, param, reset_failcount=True):
        """
        process initialization parameters

        :param param: dict of initialization parameters
        :type param: dict
        :return: nothing
        """
        verify = getParam(param, "verify", optional=True)
        if not verify:
            if getParam(param, "dynamic_phone", optional):
                self.add_tokeninfo("dynamic_phone", True)
            else:
                # specific - phone
                phone = getParam(param, "phone", required)
                self.add_tokeninfo("phone", phone)

            # in case of the sms token, only the server must know the otpkey
            # thus if none is provided, we let create one (in the TokenClass)
            if "genkey" not in param and "otpkey" not in param:
                param['genkey'] = 1

        HotpTokenClass.update(self, param, reset_failcount)
        return

    @log_with(log)
    def is_challenge_request(self, passw, user=None, options=None):
        """
        check, if the request would start a challenge

        We need to define the function again, to get rid of the
        is_challenge_request-decorator of the HOTP-Token

        :param passw: password, which might be pin or pin+otp
        :param options: dictionary of additional request parameters

        :return: returns true or false
        """
        return self.check_pin(passw, user=user, options=options)

    @log_with(log)
    def create_challenge(self, transactionid=None, options=None):
        """
        create a challenge, which is submitted to the user

        :param transactionid: the id of this challenge
        :param options: the request context parameters / data
                You can pass exception=1 to raise an exception, if
                the SMS could not be sent.
        :return: tuple of (bool, message and data)
                 bool, if submit was successful
                 message is submitted to the user
                 data is preserved in the challenge
                 reply_dict - additional reply_dict, which is added to the response
        """
        success = False
        options = options or {}
        return_message = get_action_values_from_options(SCOPE.AUTH,
                                                        "{0!s}_{1!s}".format(self.get_class_type(),
                                                                             ACTION.CHALLENGETEXT),
                                                        options) or _("Enter the OTP from the SMS:")

        return_message = return_message.replace(r'\,', ',')

        reply_dict = {'attributes': {'state': transactionid}}
        validity = self._get_sms_timeout()

        if self.is_active() is True:
            counter = self.get_otp_count()
            log.debug("counter={0!r}".format(counter))
            # At this point we must not bail out in case of a
            # Gateway error, since checkPIN is successful. A bailout
            # would cancel the checking of the other tokens
            try:
                data = None
                # Only if this is NOT a multichallenge enrollment, we try to send the sms
                if options.get("session") != CHALLENGE_SESSION.ENROLLMENT:
                    self.inc_otp_counter(counter, reset=False)
                    message_template = self._get_sms_text(options)
                    success, sent_message = self._send_sms(
                        message=message_template, options=options)

                    # Create the challenge in the database
                    if is_true(get_from_config("sms.concurrent_challenges")):
                        data = self.get_otp()[2]
                db_challenge = Challenge(self.token.serial,
                                         transaction_id=transactionid,
                                         challenge=options.get("challenge"),
                                         data=data,
                                         session=options.get("session"),
                                         validitytime=validity)
                db_challenge.save()
                transactionid = transactionid or db_challenge.transaction_id
            except Exception as e:
                info = _("The PIN was correct, but the SMS could not be sent!")
                log.warning(info + " ({0!r})".format(e))
                log.debug("{0!s}".format(traceback.format_exc()))
                return_message = info
                if is_true(options.get("exception")):
                    raise Exception(info)

        expiry_date = datetime.datetime.now() + datetime.timedelta(seconds=validity)
        reply_dict['attributes']['valid_until'] = "{0!s}".format(expiry_date)

        return success, return_message, transactionid, reply_dict

    @log_with(log)
    @check_token_locked
    def check_otp(self, anOtpVal, counter=None, window=None, options=None):
        """
        check the otpval of a token against a given counter
        and the window

        :param anOtpVal: the to be verified passw/pin
        :type anOtpVal: string

        :return: counter if found, -1 if not found
        :rtype: int
        """
        options = options or {}

        ret = HotpTokenClass.check_otp(self, anOtpVal, counter, window, options)
        if ret < 0 and is_true(get_from_config("sms.concurrent_challenges")):
            if safe_compare(options.get("data"), anOtpVal):
                # We authenticate from the saved challenge
                ret = 1
        if ret >= 0 and self._get_auto_sms(options):
            # get message template from user specific policies
            message = self._get_sms_text(options)
            self.inc_otp_counter(ret, reset=False)
            success, message = self._send_sms(message=message, options=options)
            log.debug("AutoSMS: send new SMS: {0!s}".format(success))
            log.debug("AutoSMS: {0!r}".format(message))
        return ret

    @log_with(log)
    def _send_sms(self, message="<otp>", options=None):
        """
        send sms

        :param message: the sms submit message - could contain placeholders
            like <otp> or <serial>
        :type message: string
        :param options: Additional options from the request
        :type options: dict

        :return: submitted message
        :rtype: string
        """
        if is_true(self.get_tokeninfo("dynamic_phone")):
            phone = self.user.get_user_phone("mobile")
            if isinstance(phone, list) and phone:
                # if there is a non-empty list, we use the first entry
                phone = phone[0]
        else:
            phone = self.get_tokeninfo("phone")
        if not phone:  # pragma: no cover
            log.warning("Token {0!s} does not have a phone number!".format(self.token.serial))
        otp = self.get_otp()[2]
        serial = self.get_serial()
        User = options.get("user")

        log.debug(r"sending SMS with template text: {0!s}".format(message))
        message = message.replace("<otp>", otp)
        message = message.replace("<serial>", serial)
        tags = create_tag_dict(serial=serial,
                               tokenowner=User,
                               tokentype="sms",
                               recipient={"givenname": User.info.get("givenname") if User else "",
                                          "surname": User.info.get("surname") if User else ""},
                               challenge=options.get("challenge"))
        message = message.format(otp=otp, **tags)

        # First we try to get the new SMS gateway config style
        # The token specific identifier has priority over the system-wide identifier
        sms_gateway_identifier = self.get_tokeninfo("sms.identifier") or get_from_config("sms.identifier")

        if sms_gateway_identifier:
            # New style
            sms = create_sms_instance(sms_gateway_identifier)

        else:
            # Old style
            (SMSProvider, SMSProviderClass) = self._get_sms_provider()
            log.debug("smsprovider: {0!s}, class: {1!s}".format(SMSProvider,
                                                                SMSProviderClass))

            try:
                sms = get_sms_provider_class(SMSProvider, SMSProviderClass)()
            except Exception as exc:
                log.error("Failed to load SMSProvider: {0!r}".format(exc))
                log.debug("{0!s}".format(traceback.format_exc()))
                raise exc

            try:
                # now we need the config from the env
                log.debug("loading SMS configuration for class {0!s}".format(sms))
                config = self._get_sms_provider_config()
                log.debug("config: {0!r}".format(config))
                sms.load_config(config)
            except Exception as exc:
                log.error("Failed to load sms.providerConfig: {0!r}".format(exc))
                log.debug("{0!s}".format(traceback.format_exc()))
                raise Exception("Failed to load sms.providerConfig: {0!r}".format(exc))

        log.debug("submitMessage: {0!r}, to phone {1!r}".format(message, phone))
        ret = sms.submit_message(phone, message)
        return ret, message

    @staticmethod
    @log_with(log)
    def _get_sms_provider():
        """
        get the SMS Provider class definition

        :return: tuple of SMSProvider and Provider Class as string
        :rtype: tuple of (string, string)
        """
        smsProvider = get_from_config("sms.provider",
                                      default="privacyidea.lib.smsprovider."
                                              "HttpSMSProvider.HttpSMSProvider")
        (SMSProvider, SMSProviderClass) = smsProvider.rsplit(".", 1)
        return SMSProvider, SMSProviderClass

    @staticmethod
    @log_with(log)
    def _get_sms_provider_config():
        """
        load the defined sms provider config definition

        :return: dict of the sms provider definition
        :rtype: dict
        """
        tConfig = get_from_config("sms.providerConfig", "{}")
        config = loads(tConfig)
        return config

    @staticmethod
    @log_with(log)
    def _get_sms_timeout():
        """
        get the challenge time is in the specified range

        :return: the defined validation timeout in seconds
        :rtype:  int
        """
        try:
            timeout = int(get_from_config("sms.providerTimeout", 5 * 60))
        except Exception as ex:  # pragma: no cover
            log.warning("SMSProviderTimeout: value error {0!r} - reset to 5*60".format(ex))
            timeout = 5 * 60
        return timeout

    @staticmethod
    def _get_sms_text(options):
        """
        This returns the SMSTEXT from the policy "smstext"

        options contains data like clientip, g, user and also the Request
        parameters like "challenge" or "pass".

        :param options: contains user and g object.
        :type options: dict
        :return: Message template
        :rtype: basestring
        """
        message = "<otp>"
        g = options.get("g")
        user_object = options.get("user")
        if g:
            messages = Match.user(g, scope=SCOPE.AUTH, action=SMSACTION.SMSTEXT,
                                  user_object=user_object if user_object else None).action_values(
                allow_white_space_in_action=True, unique=True)
            if len(messages) == 1:
                message = list(messages)[0]

        message = message.replace(r'\,', ',')

        return message

    @staticmethod
    def _get_auto_sms(options):
        """
        This returns the AUTOSMS setting.

        :param options: contains user and g object.
        :type options: dict
        :return: True if an SMS should be sent automatically
        :rtype: bool
        """
        autosms = False
        g = options.get("g")
        user_object = options.get("user")
        if g:
            autosmspol = Match.user(g, scope=SCOPE.AUTH, action=SMSACTION.SMSAUTO, user_object=user_object).policies()
            autosms = len(autosmspol) >= 1

        return autosms

    def prepare_verify_enrollment(self, options=None):
        """
        This is called, if the token should be enrolled in a way, that the user
        needs to provide a proof, that the server can verify, that the token
        was successfully enrolled.
        The email token needs to send an email with OTP.

        The returned dictionary is added to the response in "detail" -> "verify".

        :return: A dictionary with information that is needed to trigger the verification.
        """
        # TODO: how should we handle errors when sending the SMS?
        #  Currently we just return the message without checking the result
        _res, msg, _tid, _rdict = self.create_challenge(options=options)
        return {"message": msg}

    @classmethod
    def enroll_via_validate(cls, g, content, user_obj, message=None):
        """
        This class method is used in the policy ENROLL_VIA_MULTICHALLENGE.
        It enrolls a new token of this type and returns the necessary information
        to the client by modifying the content.

        :param g: context object
        :param content: The content of a response
        :param user_obj: A user object
        :param message: An alternative message displayed to the user during enrollment
        :return: None, the content is modified
        """
        from privacyidea.lib.token import init_token
        from privacyidea.lib.tokenclass import CLIENTMODE
        token_obj = init_token({"type": cls.get_class_type(),
                                "dynamic_phone": 1}, user=user_obj)
        content.get("result")["value"] = False
        content.get("result")["authentication"] = "CHALLENGE"

        detail = content.setdefault("detail", {})
        # Create a challenge!
        options = {"session": CHALLENGE_SESSION.ENROLLMENT, "g": g, "user": user_obj}
        c = token_obj.create_challenge(options=options)
        detail["transaction_ids"] = [c[2]]
        chal = {"transaction_id": c[2],
                "image": None,
                "client_mode": CLIENTMODE.INTERACTIVE,
                "serial": token_obj.token.serial,
                "type": token_obj.type,
                "message": message or _("Please enter your new phone number!")}
        detail["multi_challenge"] = [chal]
        detail.update(chal)

    def enroll_via_validate_2nd_step(self, passw, options=None):
        """
        This method is the optional second step of ENROLL_VIA_MULTICHALLENGE.
        It is used in situations like the email token, sms token or push,
        when enrollment via challenge response needs two steps.

        The passw is entered during the first authentication step and it
        contains the email address.

        So we need to update the token with the email address and
        we need to create a new challenge for the final authentication.

        :param options:
        :return:
        """
        self.del_tokeninfo("dynamic_phone")
        self.add_tokeninfo("phone", passw)
        # Dynamically we remember that we need to do another challenge
        self.currently_in_challenge = True



# SPDX-FileCopyrightText: (C) 2024 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-FileCopyrightText: (C) 2017 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
This script searches tokens that have not been used for a while.
To do so, it checks for the tokeninfo field "last_auth".
If the last authentication is to old, the script can delete, disable or mark
these tokens.

You can call the script like this:

    privacyidea-get-unused-tokens list 10h|7d|2y
    privacyidea-get-unused-tokens disable 10h|7d|2y
    privacyidea-get-unused-tokens delete 10h|7d|2y
    privacyidea-get-unused-tokens mark 10h|7d|2y --description="new
    value"

"""
import click
from flask.cli import with_appcontext
from privacyidea.lib.utils import get_version_number
from privacyidea.lib.token import get_tokens, remove_token, enable_token
from privacyidea.lib.policy import ACTION
from privacyidea.cli import NoPluginsFlaskGroup, create_silent_app

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(cls=NoPluginsFlaskGroup, create_app=create_silent_app,
             context_settings=CONTEXT_SETTINGS, add_default_commands=False,
             epilog='Check out our docs at https://privacyidea.readthedocs.io/ for more details')
def cli():
    """
\b
             _                    _______  _______
   ___  ____(_)  _____ _______ __/  _/ _ \\/ __/ _ |
  / _ \\/ __/ / |/ / _ `/ __/ // // // // / _// __ |
 / .__/_/ /_/|___/\\_,_/\\__/\\_, /___/____/___/_/ |_|  Get unused tokens
/_/                       /___/

   Search for tokens that have not been used for a while"""
    click.echo(r"""
                 _                    _______  _______
       ___  ____(_)  _____ _______ __/  _/ _ \/ __/ _ |
      / _ \/ __/ / |/ / _ `/ __/ // // // // / _// __ |
     / .__/_/ /_/|___/\_,_/\__/\_, /___/____/___/_/ |_|  Get unused tokens
    /_/                       /___/
    {0!s:>51}
        """.format('v{0!s}'.format(get_version_number())))


def _get_tokenlist(age):
    tlist = []
    tokenobj_list = get_tokens()
    try:
        for token_obj in tokenobj_list:
            if not token_obj.check_last_auth_newer(age):
                tlist.append(token_obj)
    except TypeError as e:
        click.secho(str(e), fg="red")

    return tlist


@click.command()
@click.argument('age')
@with_appcontext
def list_tokens(age):
    """
    Find all tokens where the last_auth is greater than AGE.

    AGE can be a value like 10h, 7d or 2y.
    """
    # TODO: bad performance, we have to look at *ALL* tokens!
    #       We can use get_tokens_paginate() here
    tlist = _get_tokenlist(age)
    if tlist:
        click.echo("Token serial\tLast authentication")
        click.echo("=" * 50)
        for token_obj in tlist:
            click.echo(f"{token_obj.token.serial}\t{token_obj.get_tokeninfo(ACTION.LASTAUTH)}")


@click.command()
@click.argument('age')
@click.option('-d', '--description', help='The description that will be set.')
@click.option('-t', '--tokeninfo',
              help='The tokeninfo that will be set. It needs a key and a '
                   'value and should be specified like key=value.')
@with_appcontext
def mark(age, description=None, tokeninfo=None):
    """
    Find unused tokens and mark them.
    They can be marked either by setting a description or by
    setting a tokeninfo.

    AGE can be a value like 10h, 7d or 2y.
    """
    tlist = _get_tokenlist(age)
    for token_obj in tlist:
        if description:
            click.echo("Setting description for token {0!s}: {1!s}".format(
                token_obj.token.serial, description))
            token_obj.set_description(description)
            token_obj.save()
        if tokeninfo:
            key, value = tokeninfo.split("=")
            click.echo("Setting tokeninfo for token {0!s}: {1!s}={2!s}".format(
                token_obj.token.serial, key, value))
            token_obj.add_tokeninfo(key, value)
            token_obj.save()


@click.command()
@click.argument('age')
@with_appcontext
def delete(age):
    """
    Find unused tokens and delete them.

    AGE can be a value like 10h, 7d or 2y.
    """
    tlist = _get_tokenlist(age)
    for token_obj in tlist:
        serial = token_obj.token.serial
        remove_token(serial)
        click.echo("Token {0!s} deleted.".format(serial))


@click.command()
@click.argument('age')
@with_appcontext
def disable(age):
    """
    Find unused tokens and disable them.

    AGE can be a value like 10h, 7d or 2y.
    """
    tlist = _get_tokenlist(age)
    for token_obj in tlist:
        serial = token_obj.token.serial
        enable_token(serial, enable=False)
        click.echo("Token {0!s} disabled.".format(serial))


cli.add_command(list_tokens, "list")
cli.add_command(disable)
cli.add_command(delete)
cli.add_command(mark)

if __name__ == '__main__':
    cli()



#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2020-09-22 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Use Multi-Challenge to ask for several tokens one after
#             the other.
#  2018-01-21 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add tokenkind
#  2015-08-28 Initial writeup of the 4eyes token
#             according to
#             https://github.com/privacyidea/privacyidea/issues/167
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This is the implementation of the 4eyes token.
The 4eyes token combines several other tokens to a virtual new token,
requiring that 2 or more users with different tokens are present to
authenticate.

A 4eyes token stores the required number of tokens of each realm
and the splitting sign.

The code is tested in tests/test_lib_tokens_4eyes.
"""
import logging
from privacyidea.api.lib.utils import getParam
from privacyidea.lib.config import get_from_config
from privacyidea.lib.log import log_with
from privacyidea.lib.tokenclass import TokenClass, TOKENKIND
from privacyidea.lib.error import ParameterError
from privacyidea.lib.token import check_realm_pass
from privacyidea.lib.decorators import check_token_locked
from privacyidea.lib import _
from privacyidea.lib.policy import ACTION, SCOPE, GROUP, get_action_values_from_options
from privacyidea.lib.challenge import get_challenges, Challenge
import json
import datetime

log = logging.getLogger(__name__)
optional = True
required = False


class FourEyesTokenClass(TokenClass):
    """
    The FourEyes token can be used to implement the Two Man Rule.
    The FourEyes token defines how many tokens of which realms are required
    like:

        * 2 tokens of RealmA
        * 1 token of RealmB

    Then users (the owners of those tokens) need to login by everyone
    entering their OTP PIN and OTP value. It does not matter, in which order
    they enter the values. All their PINs and OTPs are concatenated into one
    password field but need to be separated by the splitting sign.

    The FourEyes token again splits the password value and tries to
    authenticate each of the these passwords in the realms using the function
    ``check_realm_pass``.

    The FourEyes token itself does not provide an OTP PIN.

    The token is initialized using additional parameters at token/init:

    **Example Authentication Request**:

        .. sourcecode:: http

           POST /auth HTTP/1.1
           Host: example.com
           Accept: application/json

           type=4eyes
           user=cornelius
           realm=realm1
           4eyes=realm1:2,realm2:1
           separator=%20
    """

    def __init__(self, db_token):
        """
        :param db_token: the token
        :type db_token: database token object
        """
        TokenClass.__init__(self, db_token)
        self.set_type("4eyes")

    @staticmethod
    def get_class_type():
        """
        return the class type identifier
        """
        return "4eyes"

    @staticmethod
    def get_class_prefix():
        """
        return the token type prefix
        """
        return "PI4E"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        returns a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dict or scalar
        """
        res = {'type': '4eyes',
               'title': '4Eyes Token',
               'description': _('4Eyes Token: Use tokens of two or more users '
                                'to authenticate'),
               'init': {},
               'config': {},
               'user':  [],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin"],
               'policy': {
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of 4eyes tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of active 4eyes tokens assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               },
               }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res
        return ret

    @staticmethod
    def realms_dict_to_string(realms):
        """
        This function converts the realms - if it is a dictionary - to a string::

            {"realm1": {"selected": True,
                        "count": 1 },
             "realm2": {"selected": True,
                        "count": 2}}
                                        -> "realm1:1,realm2:2"

        :param realms: the realms as they are passed from the WebUI
        :type realms: dict
        :return: realms
        :rtype: str
        """
        realms_string = ""
        if type(realms) is dict:
            for realmname, v in realms.items():
                if v.get("selected"):
                    realms_string += "{0!s}:{1!s},".format(realmname, v.get("count"))
            if realms_string[-1] == ',':
                realms_string = realms_string[:-1]
        else:
            realms_string = realms

        return realms_string

    @staticmethod
    def convert_realms(realms):
        """
        This function converts the realms as given by the API parameter to a
        dictionary::

            "realm1:2,realm2:1" -> {"realm1":2,
                                    "realm2":1}

        :param realms: a serialized list of realms
        :type realms: str
        :return: dict of realms
        :rtype: dict
        """
        realms_dict = {}
        realm_list = realms.split(",")
        for rl in realm_list:
            r = rl.split(":")
            if len(r) == 2:
                realms_dict[r[0]] = int(r[1])
        return realms_dict

    def _get_realms(self):
        """
        This returns the dictionary how many tokens of each realm are necessary
        :return: dict with realms
        """
        return self.convert_realms(self.get_tokeninfo("4eyes"))

    @staticmethod
    def _dict_diff(required, used):
        """
        Subtract the "used" dict from the "required" dict
        :return: diff dict
        """
        result = {}
        for k, v in required.items():
            result[k] = v - used.get(k, 0)
        return result

    def _get_separator(self):
        return self.get_tokeninfo("separator") or " "

    def update(self, param):
        """
        This method is called during the initialization process.
        :param param: parameters from the token init
        :type param: dict
        :return: None
        """
        TokenClass.update(self, param)

        realms = getParam(param, "4eyes", required)
        separator = getParam(param, "separator", optional, default=" ")
        if len(separator) > 1:
            raise ParameterError("The separator must only be one single "
                                 "character")
        realms = self.realms_dict_to_string(realms)
        self.convert_realms(realms)
        self.add_tokeninfo("separator", separator)
        self.add_tokeninfo("4eyes", realms)
        self.add_tokeninfo("tokenkind", TOKENKIND.VIRTUAL)

    def _authenticate_in_realm(self, realm, password):
        """
        This method tries to authenticate a given credential (can be PIN + OTP)
        against tokens in a realm and returns the serial of the token, if successful.
        Otherwise None

        :param realm: The realm where to authenticate
        :param password: The PIN + OTP
        :return: token_id or None
        """
        serial = None
        res, reply = check_realm_pass(realm, password,
                                      exclude_types=[self.get_tokentype()])
        if res:
            serial = reply.get("serial")
        return serial

    def _authenticate_remaining_realms(self, passw, remaining_realms, used_tokens, options):
        r_success = -1
        for realm in remaining_realms:
            # check for token in realm
            serial = self._authenticate_in_realm(realm, passw)
            if serial:
                # check that not the same token is used again
                if serial in used_tokens.get(realm, []):
                    log.info("The same token {0!s} was already used. "
                             "You can not use a token twice.".format(serial))
                else:
                    # Add the serial to the used tokens.
                    if realm in used_tokens:
                        used_tokens[realm].append(serial)
                    else:
                        used_tokens[realm] = [serial]
                    options["data"] = used_tokens
                    log.debug("Partially authenticated with token {0!s}.".format(serial))
                    r_success = 1
                    break
        return r_success

    @log_with(log)
    @check_token_locked
    def authenticate(self, passw, user=None, options=None):
        """
        do the authentication on base of password / otp and user and
        options, the request parameters.

        Here we contact the other privacyIDEA server to validate the OtpVal.

        :param passw: the password / otp
        :param user: the requesting user
        :param options: the additional request parameters

        :return: tuple of (success, otp_count - 0 or -1, reply)

        """
        pin_match = True
        otp_counter = -1
        reply = None

        required_realms = self._get_realms()
        # This holds the found serial numbers in the realms
        found_serials = {}

        separator = self._get_separator()
        passwords = passw.split(separator)

        for realm in required_realms.keys():
            found_serials[realm] = []
            for otp in passwords:
                serial = self._authenticate_in_realm(realm, otp)
                if serial:
                    found_serials[realm].append(serial)
            # uniquify the serials in the list
            found_serials[realm] = list(set(found_serials[realm]))

            if len(found_serials[realm]) < required_realms[realm]:
                reply = {"foureyes": "Only found {0:d} tokens in realm {1!s}".format(
                    len(found_serials[realm]), realm)}
                otp_counter = -1
                break
            else:
                otp_counter = 1

        return pin_match, otp_counter, reply

    def _get_remaining_tokens(self, used_tokens_with_serials):
        """
        Takes a dictionary like::

            {"realm1": ["serial1", "serial2"],
             "realm3": ["serialB"]}

        and returns a dictionary of the tokens that remain for authentication like::

            {"realm1": 1, "realm3": 0}

        :param used_tokens_with_serials: dictionary with used serials
        :type used_tokens_with_serials: dict
        :return: dict
        """
        required_tokens = self._get_realms()
        used_tokens = {k: len(v) for (k, v) in used_tokens_with_serials.items()}
        # compare the dict of required tokens with the dict of used tokens
        remaining_tokens = self._dict_diff(required_tokens, used_tokens)
        return remaining_tokens

    def _get_remaining_realms(self, used_tokens_with_serials):
        """
        Return the list of the remaining realms, from which an authentication is needed.
        :param used_tokens_with_serials:
        :return:
        """
        remaining_tokens = self._get_remaining_tokens(used_tokens_with_serials)
        remaining_realms = [r for (r, toks) in remaining_tokens.items() if toks > 0]
        return remaining_realms

    @log_with(log)
    def has_further_challenge(self, options=None):
        """
        Check if there are still more tokens to be authenticated
        :param options: Options dict
        :return: True, if further challenge is required.
        """
        transaction_id = options.get('transaction_id')
        challengeobject_list = get_challenges(serial=self.token.serial,
                                              transaction_id=transaction_id)
        if len(challengeobject_list) == 1:
            remaining_realms = self._get_remaining_realms(options.get("data", {}))
            if remaining_realms:
                options["data"] = json.dumps(options.get("data", {}))
                options["message"] = "Remaining tokens: {0!s}".format(remaining_realms)
                return True
        return False

    @check_token_locked
    def check_challenge_response(self, user=None, passw=None, options=None):
        """
        This method verifies if the given response is the PIN + OTP of one of the
        remaining tokens.
        In case of success it then returns ``1``

        :param user: the requesting user
        :type user: User object
        :param passw: the password: PIN + OTP
        :type passw: string
        :param options: additional arguments from the request, which could
                        be token specific. Usually "transaction_id"
        :type options: dict
        :return: return 1 if the answer to the challenge is correct, -1 otherwise.
        :rtype: int
        """
        options = options or {}
        r_success = -1

        # fetch the transaction_id
        transaction_id = options.get('transaction_id')
        if transaction_id is None:
            transaction_id = options.get('state')

        # get the challenges for this transaction ID
        if transaction_id is not None:
            challengeobject_list = get_challenges(serial=self.token.serial,
                                                  transaction_id=transaction_id)

            for challengeobject in challengeobject_list:
                if challengeobject.is_valid():
                    # challenge is still valid
                    used_tokens = json.loads(challengeobject_list[0].data or json.dumps({}))
                    remaining_realms = self._get_remaining_realms(used_tokens)
                    r_success = self._authenticate_remaining_realms(passw, remaining_realms, used_tokens, options)

                    if r_success:
                        challengeobject.set_otp_status(True)
                    if not r_success:
                        # increase the received_count
                        challengeobject.set_otp_status()

        self.challenge_janitor()
        return r_success

    def create_challenge(self, transactionid=None, options=None):
        """
        This method creates a challenge, which is submitted to the user.
        The submitted challenge will be preserved in the challenge
        database.

        If no transaction id is given, the system will create a transaction
        id and return it, so that the response can refer to this transaction.

        :param transactionid: the id of this challenge
        :param options: the request context parameters / data
        :type options: dict
        :return: tuple of (bool, message, transactionid, reply_dict)
        :rtype: tuple

        The return tuple builds up like this:
        ``bool`` if submit was successful;
        ``message`` which is displayed in the JSON response;
        additional challenge ``reply_dict``, which are displayed in the JSON challenges response.
        """
        options = options or {}
        message = ""
        if type(options.get("data")) == dict:
            # In the special first chal-resp case we do not have jsonified data, yet. So we need to convert
            options["data"] = json.dumps(options.get("data"))
        used_tokens = json.loads(options.get("data", json.dumps({})))
        remaining_realms = self._get_remaining_realms(used_tokens)
        if remaining_realms:
            message = "Please authenticate with another token from " \
                      "either realm: {0!s}.".format(", ".join(remaining_realms))

        validity = int(get_from_config('DefaultChallengeValidityTime', 120))
        tokentype = self.get_tokentype().lower()
        # Maybe there is a 4EYESChallengeValidityTime...
        lookup_for = tokentype.capitalize() + 'ChallengeValidityTime'
        validity = int(get_from_config(lookup_for, validity))

        # Create the challenge in the database
        db_challenge = Challenge(self.token.serial,
                                 transaction_id=transactionid,
                                 data=options.get("data"),
                                 session=options.get("session"),
                                 challenge=message,
                                 validitytime=validity)
        db_challenge.save()
        expiry_date = datetime.datetime.now() + \
                      datetime.timedelta(seconds=validity)
        reply_dict = {'attributes': {'valid_until': "{0!s}".format(expiry_date)}}
        return True, message, db_challenge.transaction_id, reply_dict

    def is_challenge_request(self, passw, user=None, options=None):
        """
        The 4eyes token can act as a challenge response token.

        Either
         * if the first passw given is the PIN of the 4eyes token or
         * if the first passw given is the complete PIN+OTP from one of
           the admintokens.

        :param passw: password, which might be pin or pin+otp
        :type passw: str
        :param user: The user from the authentication request
        :type user: User object
        :param options: dictionary of additional request parameters
        :type options: dict
        :return: true or false
        :rtype: bool
        """
        options = options or {}
        pin_match = self.check_pin(passw, user=user, options=options)
        if pin_match:
            return True
        elif "transaction_id" not in options:
            # The special case, when the *first* admin token already matches, we will
            # start a multi challenge chain
            used_tokens = {}
            remaining_realms = self._get_remaining_realms({})
            r_success = self._authenticate_remaining_realms(passw, remaining_realms, used_tokens, options)
            return r_success >= 0

        return False



"""v3.10: Add columns authentication, user_agent/version to Audit

Revision ID: 250931d82e51
Revises: 5741e5dac477
Create Date: 2024-03-27 16:08:47.242337

"""

# revision identifiers, used by Alembic.
revision = '250931d82e51'
down_revision = '5741e5dac477'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.add_column('pidea_audit', sa.Column('authentication', sa.Unicode(length=12), nullable=True))
        op.add_column('pidea_audit', sa.Column('user_agent', sa.Unicode(length=20), nullable=True))
        op.add_column('pidea_audit', sa.Column('user_agent_version',
                                               sa.Unicode(length=20), nullable=True))
    except Exception as exx:
        print("Can not add columns 'authentication', 'user_agent', 'user_agent_version' to "
              "table 'pidea_audit'. Probably already exist.")
        print(exx)
        raise


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pidea_audit', 'user_agent_version')
    op.drop_column('pidea_audit', 'user_agent')
    op.drop_column('pidea_audit', 'authentication')
    # ### end Alembic commands ###



import sys
sys.stdout = sys.stderr
from privacyidea.app import create_app
# Now we can select the config file:
application = create_app(config_name="production", config_file="/etc/privacyidea/pi.cfg")



"""
This testcase is used to test the REST API  in api/machines.py
to fetch machine information and to attach token to machines
"""
import passlib

from privacyidea.lib.user import User
from .base import MyApiTestCase
from privacyidea.lib.token import init_token, get_tokens, remove_token
from privacyidea.lib.machine import attach_token, detach_token, ANY_MACHINE, NO_RESOLVER
from privacyidea.lib.policy import (set_policy, delete_policy, ACTION, SCOPE)

HOSTSFILE = "tests/testdata/hosts"

SSHKEY = "ssh-rsa " \
         "AAAAB3NzaC1yc2EAAAADAQABAAACAQDJy0rLoxqc8SsY8DVAFijMsQyCv" \
         "hBu4K40hdZOacXK4O6OgnacnSKN56MP6pzz2+4svzvDzwvkFsvf34pbsgD" \
         "F67PPSCsimmjEQjf0UfamBKh0cl181CbPYsph3UTBOCgHh3FFDXBduPK4DQz" \
         "EVQpmqe80h+lsvQ81qPYagbRW6fpd0uWn9H7a/qiLQZsiKLL07HGB+NwWue4os" \
         "0r9s4qxeG76K6QM7nZKyC0KRAz7CjAf+0X7YzCOu2pzyxVdj/T+KArFcMmq8V" \
         "dz24mhcFFXTzU3wveas1A9rwamYWB+Spuohh/OrK3wDsrryStKQv7yofgnPMs" \
         "TdaL7XxyQVPCmh2jVl5ro9BPIjTXsre9EUxZYFVr3EIECRDNWy3xEnUHk7Rzs" \
         "734Rp6XxGSzcSLSju8/MBzUVe35iXfXDRcqTcoA0700pIb1ANYrPUO8Up05v4" \
         "EjIyBeU61b4ilJ3PNcEVld6FHwP3Z7F068ef4DXEC/d7pibrp4Up61WYQIXV/" \
         "utDt3NDg/Zf3iqoYcJNM/zIZx2j1kQQwqtnbGqxJMrL6LtClmeWteR4420uZx" \
         "afLE9AtAL4nnMPuubC87L0wJ88un9teza/N02KJMHy01Yz3iJKt3Ou9eV6kqO" \
         "ei3kvLs5dXmriTHp6g9whtnN6/Liv9SzZPJTs8YfThi34Wccrw== " \
         "NetKnights GmbH"
SSHKEY_ecdsa = "ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzd" \
               "HAyNTYAAABBBHGCdIk0pO1HFr/mF4oLb43ZRyQJ4K7ICLrAhAiQERVa0tUvyY5TE" \
               "zurWTqxSMx203rY77t6xnHLZBMPPpv8rk0= cornelius@puck"
OTPKEY = "3132333435363738393031323334353637383930"


class APIMachinesTestCase(MyApiTestCase):

    serial2 = "ser1"
    serial3 = "UBOM12345"
    serial4 = "OATH1234"

    def test_00_create_machine_resolver(self):
        # create a machine resolver
        with self.app.test_request_context('/machineresolver/machineresolver1',
                                           data={'type': 'hosts',
                                                 'filename': HOSTSFILE},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"] is True, result)
            self.assertTrue(result["value"] == 1, result)

    def test_01_get_machine_list(self):
        with self.app.test_request_context('/machine/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertEqual(len(result["value"]), 5)
            self.assertTrue("hostname" in result["value"][0])
            self.assertTrue("id" in result["value"][0])
            self.assertTrue("ip" in result["value"][0])
            self.assertTrue("resolver_name" in result["value"][0])

    def test_01_get_machine_list_any(self):
        with self.app.test_request_context('/machine/?any=192',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertEqual(len(result["value"]), 4)
            self.assertTrue("hostname" in result["value"][0])
            self.assertTrue("id" in result["value"][0])
            self.assertTrue("ip" in result["value"][0])
            self.assertTrue("resolver_name" in result["value"][0])

    def test_02_attach_token(self):
        serial = "S1"
        # create token
        init_token({"serial": serial, "type": "spass"})

        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"hostname": "gandalf",
                                                 "serial": serial,
                                                 "application": "luks",
                                                 "slot": "1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[0].application, "luks")
        self.assertEqual(token_obj.token.machine_list[0].option_list[0].mt_key,
                         "slot")

    def test_03_attach_offline_token(self):
        # The offline token allows to be attached without a machine.
        # create token
        serial = "offHOTP"
        init_token({"serial": serial, "genkey": 1})

        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"resolver": "",
                                                 "machineid": 0,
                                                 "serial": serial,
                                                 "application": "offline",
                                                 "count": "12"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[0].application, "offline")
        self.assertEqual(token_obj.token.machine_list[0].option_list[0].mt_key,
                         "count")

        # Get the token
        with self.app.test_request_context('/machine/token',
                                           method='GET',
                                           query_string={"serial": serial},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            machine_list = result.get("value")
            self.assertEqual(len(machine_list), 1)
            self.assertEqual(machine_list[0].get("serial"), serial)
            self.assertEqual(machine_list[0].get("hostname"), "any host")

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[0].application, "offline")
        self.assertEqual(token_obj.token.machine_list[0].option_list[0].mt_key,
                         "count")

        # Now detach the offline token. In this case we ignore the machine and resolver.
        with self.app.test_request_context('/machine/token/{0!s}/{1!s}/{2!s}/offline'.format(serial,
                                                                                             ANY_MACHINE,
                                                                                             NO_RESOLVER),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertEqual(result["value"], 1)

        # check that the token has no applications/machines anymore
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(len(token_obj.token.machine_list), 0)

        remove_token(serial)

    def test_04_set_options(self):
        serial = "S1"
        with self.app.test_request_context('/machine/tokenoption',
                                           method='POST',
                                           data={"hostname": "gandalf",
                                                 "serial": serial,
                                                 "application": "luks",
                                                 "partition": "/dev/sdb1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True, result)
            self.assertGreaterEqual(result["value"]["added"], 1, result)
            self.assertEqual(result['value']['deleted'], 0, result)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[0].application, "luks")
        self.assertEqual(token_obj.token.machine_list[0].option_list[
                             1].mt_value, "/dev/sdb1")

        # delete slot!
        with self.app.test_request_context('/machine/tokenoption',
                                           method='POST',
                                           data={"hostname": "gandalf",
                                                 "serial": serial,
                                                 "application": "luks",
                                                 "slot": ""},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True, result)
            self.assertEqual(result["value"]["added"], 0, result)
            self.assertGreaterEqual(result['value']['deleted'], 1, result)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[0].application, "luks")
        # As we deleted the slot, the partition now is the only entry in the
        # list
        self.assertEqual(token_obj.token.machine_list[0].option_list[
                             0].mt_value, "/dev/sdb1")

        # Overwrite option
        with self.app.test_request_context('/machine/tokenoption',
                                           method='POST',
                                           data={"hostname": "gandalf",
                                                 "serial": serial,
                                                 "application": "luks",
                                                 "partition": "/dev/sda1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True, result)
            self.assertGreaterEqual(result["value"]["added"], 1, result)
            self.assertEqual(result['value']['deleted'], 0, result)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[0].application, "luks")
        # As we deleted the slot, the partition now is the only entry in the
        # list
        self.assertEqual(token_obj.token.machine_list[0].option_list[
                             0].mt_value, "/dev/sda1")

    def test_04_set_options_by_mtid(self):
        serial = "S1"
        # current number of attached applications.
        token_obj = get_tokens(serial=serial)[0]
        num_applications = len(token_obj.token.machine_list)
        # create an ssh application
        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"hostname": "gandalf",
                                                 "serial": serial,
                                                 "application": "ssh",
                                                 "user": "root",
                                                 "service_id": "webserver"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)
            mtid = result.get("value")

        with self.app.test_request_context('/machine/tokenoption',
                                           method='POST',
                                           data={"mtid": mtid,
                                                 "application": "ssh",
                                                 "service_id": "mailserver",
                                                 "user": ""},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True, result)
            self.assertEqual(1, result["value"]["added"], result)
            self.assertEqual(1, result['value']['deleted'], result)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[1].application, "ssh")
        self.assertEqual(token_obj.token.machine_list[1].option_list[
                             0].mt_value, "mailserver")
        # Delete machinetoken
        with self.app.test_request_context(
                '/machine/token/S1/ssh/{}'.format(mtid),
                method='DELETE',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

        # check if the the application is detached again
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(num_applications, len(token_obj.token.machine_list))

    def test_05_list_machinetokens(self):
        with self.app.test_request_context('/machine/token?serial=S1',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertEqual(len(result["value"]), 1)
            self.assertTrue(result["value"][0]["application"] == "luks")

        with self.app.test_request_context('/machine/token?hostname=gandalf',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertEqual(len(result["value"]), 1)
            self.assertTrue(result["value"][0]["application"] == "luks")

    def test_99_detach_token(self):
        serial = "S1"
        # create token
        init_token({"serial": serial, "type": "spass"})

        # Gandalf is 192.168.0.1
        with self.app.test_request_context(
                '/machine/token/S1/192.168.0.1/machineresolver1/luks',
                method='DELETE',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(len(token_obj.token.machine_list), 0)

    def test_10_auth_items_ssh_rsa(self):
        # create an SSH token
        token_obj = init_token({"serial": self.serial2, "type": "sshkey",
                                "sshkey": SSHKEY})
        self.assertEqual(token_obj.type, "sshkey")

        # Attach the token to the machine "gandalf" with the application SSH
        r = attach_token(hostname="gandalf", serial=self.serial2,
                         application="ssh", options={"user": "testuser"})
        mtid = r.id

        self.assertEqual(r.machine_id, "192.168.0.1")

        # fetch the auth_items for application SSH on machine gandalf
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ssh-rsa"), sshkey)

        # fetch the auth_items for user testuser
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf&user=testuser',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ssh-rsa"), sshkey)

        # fetch auth_items for testuser, but with mangle policy
        # Remove everything that sounds like "SOMETHING\" in front of
        # the username
        set_policy(name="mangle1", scope=SCOPE.AUTH,
                   action="{0!s}=user/.*\\\\(.*)/\\1/".format(ACTION.MANGLE))
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf&user=DOMAIN\\testuser',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ssh-rsa"), sshkey)
        delete_policy("mangle1")

        # Now that the policy is deleted, we will not get the auth_items
        # anymore, since the username is not mangled.
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf&user=DOMAIN\\testuser',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkeys = result["value"].get("ssh")
            # No user DOMAIN\\testuser and no SSH keys
            self.assertFalse(sshkeys)

        # fetch the auth_items on machine gandalf for all applications
        with self.app.test_request_context(
                '/machine/authitem?hostname=gandalf',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ssh-rsa"), sshkey)

        # Detach the machinetoken via ID - this is used in the UI
        with self.app.test_request_context("/machine/token/{0!s}/ssh/{1!s}".format(self.serial2, mtid),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        remove_token(self.serial2)

    def test_10_auth_items_ssh_rsa_with_service_id(self):
        # Attach with service_id and without IP, but still also support IP
        # create an SSH token
        token_obj = init_token({"serial": self.serial2, "type": "sshkey",
                                "sshkey": SSHKEY})
        self.assertEqual(token_obj.type, "sshkey")

        # Attach the token to the machine "gandalf" with the application SSH
        r = attach_token(serial=self.serial2,
                         application="ssh", options={"user": "testuser", "service_id": "webserver"})

        self.assertEqual(None, r.machine_id)
        self.assertEqual("ssh", r.application)

        # fetch the auth_items for application SSH on machine gandalf
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf&service_id=webserver',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ssh-rsa"), sshkey)

        # fetch the auth_items for user testuser
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf&service_id=webserver&user=testuser',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ssh-rsa"), sshkey)

        # fetch the auth_items for a user, who has no ssh keys
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf&service_id=webserver&user=root',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertEqual({}, result.get("value"))

        # fetch auth_items for testuser, but with mangle policy
        # Remove everything that sounds like "SOMETHING\" in front of
        # the username
        set_policy(name="mangle1", scope=SCOPE.AUTH,
                   action="{0!s}=user/.*\\\\(.*)/\\1/".format(ACTION.MANGLE))
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf&service_id=webserver&user=DOMAIN\\testuser',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ssh-rsa"), sshkey)
        delete_policy("mangle1")

        # Now that the policy is deleted, we will not get the auth_items
        # anymore, since the username is not mangled.
        with self.app.test_request_context(
                '/machine/authitem/ssh?service_id=webserver&hostname=gandalft&user=DOMAIN\\testuser',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkeys = result["value"].get("ssh")
            # No user DOMAIN\\testuser and no SSH keys
            self.assertFalse(sshkeys)

        detach_token(self.serial2, application="ssh")
        remove_token(self.serial2)

    def test_10_auth_items_ssh_ecdsa(self):
        # create an SSH token
        token_obj = init_token({"serial": self.serial2, "type": "sshkey",
                                "sshkey": SSHKEY_ecdsa})
        self.assertEqual(token_obj.type, "sshkey")

        # Attach the token to the machine "gandalf" with the application SSH
        r = attach_token(hostname="gandalf", serial=self.serial2,
                         application="ssh", options={"user": "testuser"})

        self.assertEqual(r.machine_id, "192.168.0.1")

        # fetch the auth_items for application SSH on machine gandalf
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ecdsa-sha2-nistp256"), sshkey)

        # fetch the auth_items for user testuser
        with self.app.test_request_context(
                '/machine/authitem/ssh?hostname=gandalf&user=testuser',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            sshkey = result["value"].get("ssh")[0].get("sshkey")
            self.assertTrue(sshkey.startswith("ecdsa-sha2-nistp256"), sshkey)

        detach_token(self.serial2, application="ssh", hostname="gandalf")
        remove_token(self.serial2)

    def test_11_auth_items_luks(self):
        # create TOTP/Yubikey token
        token_obj = init_token({"serial": self.serial3, "type": "totp",
                                "otpkey": "12345678"})
        self.assertEqual(token_obj.type, "totp")

        # Attach the token to the machine "gandalf" with the application SSH
        r = attach_token(hostname="gandalf", serial=self.serial3,
                         application="luks", options={"slot": "1",
                                                      "partition": "/dev/sda1"})

        self.assertEqual(r.machine_id, "192.168.0.1")

        # fetch the auth_items on machine gandalf for application luks
        with self.app.test_request_context(
                '/machine/authitem/luks?hostname=gandalf',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            slot = result["value"].get("luks")[0].get("slot")
            self.assertEqual(slot, "1")

        # fetch the auth_items on machine gandalf for application luks
        with self.app.test_request_context(
                '/machine/authitem/luks?hostname=gandalf&challenge=abcdef',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            slot = result["value"].get("luks")[0].get("slot")
            self.assertEqual(slot, "1")
            response = result["value"].get("luks")[0].get("response")
            self.assertEqual(response,
                             "93235fc7d1d444d0ec014ea9eafcc44fc65b73eb")

    def test_12_auth_items_offline(self):
        # Create HOTP token for offline usage
        self.setUp_user_realms()
        token_obj = init_token({"serial": self.serial4, "type": "hotp",
                                "otpkey": OTPKEY,
                                "pin": "test"}, User("cornelius", self.realm1))
        self.assertEqual(token_obj.type, "hotp")
        self.assertEqual(token_obj.token.count, 0)

        # Attach the token to the machine "gandalf" with the application offline
        r = attach_token(hostname="gandalf", serial=self.serial4,
                         application="offline", options={"count": 17})

        self.assertEqual(r.machine_id, "192.168.0.1")

        # fetch the auth_items on machine gandalf for application "offline"
        with self.app.test_request_context(
                '/machine/authitem/offline?hostname=gandalf',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            offline_auth_item = result["value"].get("offline")[0]
            username = offline_auth_item.get("user")
            self.assertEqual(username, "cornelius")
            # check, if we got 17 otp values
            response = offline_auth_item.get("response")
            self.assertEqual(len(response), 17)
            self.assertEqual(token_obj.token.count, 17)
            self.assertTrue(passlib.hash.
                            pbkdf2_sha512.verify("755224",
                                                 response.get('0')))

        self.assertEqual(token_obj.check_otp('187581'), -1)  # count = 16
        with self.app.test_request_context(
                '/validate/check?user=cornelius&pass=test447589',  # count = 17
                environ_base={'REMOTE_ADDR': '192.168.0.1'},
                method='GET'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json
            self.assertTrue(result['result']['status'])
            self.assertTrue(result['result']['value'])
            offline_auth_item = result["auth_items"]["offline"][0]
            username = offline_auth_item.get("user")
            self.assertEqual(username, "cornelius")
            # check, if we got 17 otp values
            response = offline_auth_item.get("response")
            self.assertEqual(len(response), 17)
            self.assertEqual(token_obj.token.count, 35)  # 17 + 17 + 1, because we consumed 447589
            self.assertTrue(passlib.hash.
                            pbkdf2_sha512.verify("test903435",  # count = 18
                                                 response.get('18')))
            self.assertTrue(passlib.hash.
                            pbkdf2_sha512.verify("test749439",  # count = 34
                                                 response.get('34')))
        self.assertEqual(token_obj.check_otp('747439'), -1)  # count = 34
        self.assertEqual(token_obj.check_otp('037211'), 35)  # count = 35

    def test_20_detach_ssh_key_any_token(self):
        # we could also attach an SSH key to "any machine".
        # We need to check, that we can also detach this token.
        serial = "SSHany"
        token_obj = init_token({"serial": serial, "type": "sshkey",
                                "sshkey": SSHKEY})
        self.assertEqual(token_obj.type, "sshkey")

        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"resolver": "",
                                                 "machineid": 0,
                                                 "serial": serial,
                                                 "application": "ssh",
                                                 "user": "root"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[0].application, "ssh")
        self.assertEqual(token_obj.token.machine_list[0].option_list[0].mt_key,
                         "user")

        # Get the token
        with self.app.test_request_context('/machine/token',
                                           method='GET',
                                           query_string={"serial": serial},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            machine_list = result.get("value")
            self.assertEqual(len(machine_list), 1)
            self.assertEqual(machine_list[0].get("serial"), serial)
            self.assertEqual(machine_list[0].get("hostname"), "any host")

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual("ssh", token_obj.token.machine_list[0].application)
        self.assertEqual("user", token_obj.token.machine_list[0].option_list[0].mt_key)

        # Now detach the ssh token from any machine
        with self.app.test_request_context('/machine/token/{0!s}/any%20machine/no%20resolver/ssh'.format(serial),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertEqual(result["value"], 1)

        # check that the token has no applications/machines anymore
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(len(token_obj.token.machine_list), 0)

        remove_token(serial)

    def test_30_detach_old_offline_token(self):
        # Old offline tokens are attached to a distinct machine in a resolver.
        # We need to ensure, that the new code can also detach these old tokens.
        # 1. Create machine resolver
        # We are using the existing machine resolver: 192.168.0.1/machineresolver1 (gandalf)
        # 2. Create an HOTP token
        serial = "hotp01"
        tok = init_token({"type": "hotp", "otpkey": self.otpkey, "serial": serial}, User("cornelius", self.realm1))
        self.assertEqual("hotp", tok.token.tokentype)
        # 3. Attach this token to a distinct machine
        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"hostname": "gandalf",
                                                 "serial": serial,
                                                 "application": "offline"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"])
            self.assertTrue(result["value"] >= 1)

        # check if the options were set.
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(token_obj.token.machine_list[0].machine_id, "192.168.0.1")
        self.assertEqual(token_obj.token.machine_list[0].application, "offline")

        # 4. Authenticate
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           environ_base={'REMOTE_ADDR': '192.168.0.1'},
                                           data={"user": "cornelius",
                                                 "pass": self.valid_otp_values[0]}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)
            detail = res.json.get("detail")
            self.assertEqual(serial, detail.get("serial"))
            # check for offline data
            auth_items = res.json.get("auth_items")
            self.assertIn("offline", auth_items)
            offline = auth_items.get("offline")
            self.assertEqual(1, len(offline))
            self.assertIn("refilltoken", offline[0])

        # 5. Detach this token from the offline application and machine
        with self.app.test_request_context('/machine/token/{0!s}/192.168.0.1/machineresolver1/offline'.format(serial),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            # One recorded deleted
            self.assertEqual(result["value"], 1)

        # check that the token has no applications/machines anymore
        token_obj = get_tokens(serial=serial)[0]
        self.assertEqual(len(token_obj.token.machine_list), 0)
        remove_token(serial)

    def test_35_delete_offline_token(self):
        # Delete an offline token with machine token options (Issue #4136)
        serial = "hotp01"
        tok = init_token({"type": "hotp", "otpkey": self.otpkey, "serial": serial})
        # Mark this token as "offline"
        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"machineid": 0,
                                                 "serial": serial,
                                                 "application": "offline",
                                                 "rounds": 25,
                                                 "count": 15,
                                                 "resolver": ""},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            result = res.json.get("result")
            self.assertTrue(result["status"])
            self.assertGreaterEqual(result["value"], 1, result)

        # Check if the options were set.
        with self.app.test_request_context('/machine/token', method="GET",
                                           query_string={"serial": serial},
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            result = res.json.get("result")
            self.assertEqual("offline", result["value"][0]["application"], result)
            self.assertEqual("15", result["value"][0]["options"]["count"], result)
            self.assertEqual("25", result["value"][0]["options"]["rounds"], result)

        # Delete the token
        with self.app.test_request_context(f"/token/{serial}", method="DELETE",
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            result = res.json.get("result")
            self.assertTrue(result["status"])
            self.assertEqual(1, result["value"], result)

        # Check that the token was deleted
        self.assertFalse(get_tokens(serial=serial))



# http://www.privacyidea.org
# (c) Cornelius Kölbel
#
# 2016-02-20 Cornelius Kölbel, <cornelius@privacyidea.org>
#            Implement REST API, create, update, delete, list
#            for RADIUS server definitions
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This endpoint is used to create, update, list and delete RADIUS
server definitions. RADIUS server definitions can be used for several purposes
like RADIUS-Token or RADIUS-passthru policies.

The code of this module is tested in tests/test_api_radiusserver.py
"""
from flask import (Blueprint, request)
from .lib.utils import (getParam,
                        required,
                        send_result)
from ..lib.log import log_with
from ..lib.policy import ACTION
from ..api.lib.prepolicy import prepolicy, check_base_action
from flask import g
import logging
from privacyidea.lib.radiusserver import (add_radius, list_radiusservers,
                                          delete_radius, test_radius)


log = logging.getLogger(__name__)

radiusserver_blueprint = Blueprint('radiusserver_blueprint', __name__)


@radiusserver_blueprint.route('/<identifier>', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.RADIUSSERVERWRITE)
@log_with(log)
def create(identifier=None):
    """
    This call creates or updates a RADIUS server definition.

    :param identifier: The unique name of the RADIUS server definition
    :param server: The FQDN or IP of the RADIUS server
    :param port: The port of the RADIUS server
    :param secret: The RADIUS secret of the RADIUS server
    :param description: A description for the definition
    """
    param = request.all_data
    identifier = identifier.replace(" ", "_")
    server = getParam(param, "server", required)
    port = int(getParam(param, "port", default=1812))
    secret = getParam(param, "secret", required)
    retries = int(getParam(param, "retries", default=3))
    timeout = int(getParam(param, "timeout", default=5))
    description = getParam(param, "description", default="")
    dictionary = getParam(param, "dictionary",
                          default="/etc/privacyidea/dictionary")

    r = add_radius(identifier, server, secret, port=port,
                   description=description, dictionary=dictionary,
                   retries=retries, timeout=timeout)

    g.audit_object.log({'success': r > 0,
                        'info':  r})
    return send_result(r > 0)


@radiusserver_blueprint.route('/', methods=['GET'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.RADIUSSERVERREAD)
def list_radius():
    """
    This call gets the list of RADIUS server definitions
    """
    res = list_radiusservers()
    # We do not add the secret!
    for identifier, data in res.items():
        data.pop("password")
    g.audit_object.log({'success': True})
    return send_result(res)


@radiusserver_blueprint.route('/<identifier>', methods=['DELETE'])
@prepolicy(check_base_action, request, ACTION.RADIUSSERVERWRITE)
@log_with(log)
def delete_server(identifier=None):
    """
    This call deletes the specified RADIUS server configuration

    :param identifier: The unique name of the RADIUS server definition
    """
    r = delete_radius(identifier)

    g.audit_object.log({'success': r > 0,
                        'info':  r})
    return send_result(r > 0)


@radiusserver_blueprint.route('/test_request', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.RADIUSSERVERWRITE)
@log_with(log)
def test():
    """
    Test the RADIUS definition
    :return:
    """
    param = request.all_data
    identifier = getParam(param, "identifier", required)
    server = getParam(param, "server", required)
    port = int(getParam(param, "port", default=1812))
    secret = getParam(param, "secret", required)
    retries = int(getParam(param, "retries", default=3))
    timeout = int(getParam(param, "timeout", default=5))
    user = getParam(param, "username", required)
    password = getParam(param, "password", required)
    dictionary = getParam(param, "dictionary",
                          default="/etc/privacyidea/dictionary")

    r = test_radius(identifier, server, secret, user, password, port=port,
                    dictionary=dictionary, retries=retries, timeout=timeout)
    g.audit_object.log({'success': r > 0,
                        'info':  r})
    return send_result(r > 0)



"""Add subscription table

Revision ID: 37e6b49fc686
Revises: 3c6e9dd7fbac
Create Date: 2016-10-23 10:45:45.792467

"""

# revision identifiers, used by Alembic.
revision = '37e6b49fc686'
down_revision = '3c6e9dd7fbac'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError, InternalError


def upgrade():
    try:
        op.create_table('subscription',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('application', sa.Unicode(length=30), nullable=True),
        sa.Column('for_name', sa.Unicode(length=50), nullable=False),
        sa.Column('for_address', sa.Unicode(length=128), nullable=True),
        sa.Column('for_email', sa.Unicode(length=128), nullable=False),
        sa.Column('for_phone', sa.Unicode(length=50), nullable=False),
        sa.Column('for_url', sa.Unicode(length=80), nullable=True),
        sa.Column('for_comment', sa.Unicode(length=255), nullable=True),
        sa.Column('by_name', sa.Unicode(length=50), nullable=False),
        sa.Column('by_email', sa.Unicode(length=128), nullable=False),
        sa.Column('by_address', sa.Unicode(length=128), nullable=True),
        sa.Column('by_phone', sa.Unicode(length=50), nullable=True),
        sa.Column('by_url', sa.Unicode(length=80), nullable=True),
        sa.Column('date_from', sa.DateTime(), nullable=True),
        sa.Column('date_till', sa.DateTime(), nullable=True),
        sa.Column('num_users', sa.Integer(), nullable=True),
        sa.Column('num_tokens', sa.Integer(), nullable=True),
        sa.Column('num_clients', sa.Integer(), nullable=True),
        sa.Column('level', sa.Unicode(length=30), nullable=True),
        sa.Column('signature', sa.Unicode(length=640), nullable=True),
        sa.PrimaryKeyConstraint('id')
        )
        op.create_index(op.f('ix_subscription_application'), 'subscription', ['application'], unique=False)
        op.create_index(op.f('ix_subscription_id'), 'subscription', ['id'], unique=False)
    except (OperationalError, ProgrammingError, InternalError) as exx:
        if "duplicate column name" in str(exx.orig).lower():
            print("Good. Table subscription already exists.")
        else:
            print("Table subscription exists")
            print(exx)

    except Exception as exx:
        print("Could not add Table subscription")
        print (exx)


def downgrade():
    op.drop_index(op.f('ix_subscription_id'), table_name='subscription')
    op.drop_index(op.f('ix_subscription_application'), table_name='subscription')
    op.drop_table('subscription')

